#  _________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright (c) 2014 Sandia Corporation.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  This software is distributed under the BSD License.
#  _________________________________________________________________________

# TODO: Workaround when mpi4py is not available or COMM_WORLD is
# TODO: Add option to launch without calling MPI_Comm_spawn
# TODO: Figure out what to do with working_directory, logfile, and output_solver_log
#       when MPI_Comm_spawn is called.

import io
import os
import sys
import time
import array

import pyutilib.subprocess

from pyomo.core import ComponentUID
from pyomo.opt import (ReaderFactory,
                       ResultsFormat,
                       ProblemFormat,
                       UndefinedData)
from pyomo.pysp.util.configured_object import PySPConfiguredObject
from pyomo.pysp.util.config import (PySPConfigValue,
                                    PySPConfigBlock,
                                    safe_register_common_option,
                                    safe_register_unique_option,
                                    safe_declare_common_option,
                                    safe_declare_unique_option,
                                    _domain_must_be_str,
                                    _domain_tuple_of_str_or_dict)
from pyomo.pysp.util.misc import (parse_command_line,
                                  launch_command)
from pyomo.pysp.scenariotree.manager import \
    (InvocationType,
     ScenarioTreeManagerFactory,
     ScenarioTreeManagerClientPyro,
     ScenarioTreeSolveResults)
from pyomo.pysp.scenariotree.manager_solver import \
    ScenarioTreeManagerSolverClientPyro
from pyomo.pysp.solvers.spsolver import (SPSolverResults,
                                         SPSolverFactory)
from pyomo.pysp.solvers.spsolvershellcommand import \
    SPSolverShellCommand
from pyomo.pysp.convert.schuripopt import \
    (_write_bundle_nl,
     _write_scenario_nl,
     _write_problem_list_file)

from six.moves import xrange
# use fast version of pickle (python 2 or 3)
from six.moves import cPickle as pickle

# generate an absolute path to this file
thisfile = os.path.abspath(__file__)

_schuripopt_group_label = "SchurIpoptSolver Options"

def EXTERNAL_invoke_solve(worker,
                          working_directory,
                          subproblem_type,
                          logfile,
                          problem_list_filename,
                          executable,
                          output_solver_log,
                          io_options,
                          command_line_options,
                          options_filename,
                          suffixes=None):
    assert os.path.exists(working_directory)
    if suffixes is None:
        suffixes = [".*"]

    #
    # Write the NL files for the subproblems local to
    # this worker
    #

    filedata = {}
    write_time = {}
    load_function = None
    if subproblem_type == 'bundles':
        assert worker.scenario_tree.contains_bundles()
        load_function = worker._process_bundle_solve_result
        for bundle in worker.scenario_tree.bundles:
            start = time.time()
            filedata[bundle.name] = _write_bundle_nl(
                worker,
                bundle,
                working_directory,
                io_options)
            stop = time.time()
            write_time[bundle.name] = stop - start
    else:
        assert subproblem_type == 'scenarios'
        load_function = worker._process_scenario_solve_result
        for scenario in worker.scenario_tree.scenarios:
            start = time.time()
            filedata[scenario.name] = _write_scenario_nl(
                worker,
                scenario,
                working_directory,
                io_options)
            stop = time.time()
            write_time[scenario.name] = stop - start
    assert load_function is not None
    assert len(filedata) > 0

    args = []
    args.append(problem_list_filename)
    args.append("-AMPL")
    args.append("use_problem_file=yes")
    args.append("option_file_name="+options_filename)
    root_node = worker.scenario_tree.findRootNode()
    if hasattr(worker, "MPI"):
        args.append("mpi_spawn_mode=yes")
        if worker.mpi_comm_tree[root_node.name].rank == 0:
            args.append("output_file="+str(logfile))

    for key, val in command_line_options:
        key = key.strip()
        if key == "use_problem_file":
            raise ValueError(
                "Use of the 'use_problem_file' command-line "
                "option is disallowed.")
        elif key == "mpi_spawn_mode":
            raise ValueError(
                "Use of the 'mpi_spawn_mode' command-line "
                "option is disallowed.")
        elif key == "option_file_name":
            raise ValueError(
                "Use of the 'option_file_name' command-line "
                "option is disallowed.")
        elif key == "output_file":
            raise ValueError(
                "Use of the 'output_file' command-line "
                "option is disallowed.")
        elif key == '-AMPL':
            raise ValueError(
                "Use of the '-AMPL' command-line "
                "option is disallowed.")
        else:
            args.append(key+"="+str(val))

    start = time.time()
    if hasattr(worker, "MPI"):
        currdir = os.getcwd()
        try:
            os.chdir(working_directory)
            spawn = worker.mpi_comm_tree[root_node.name].Spawn(
                executable,
                args=args,
                maxprocs=worker.mpi_comm_tree[root_node.name].size)
            rc = None
            if worker.mpi_comm_tree[root_node.name].rank == 0:
                rc = array.array("i", [0])
                spawn.Reduce(sendbuf=None,
                             recvbuf=[rc, worker.MPI.INT],
                             op=worker.MPI.SUM,
                             root=worker.MPI.ROOT)
            rc = worker.mpi_comm_tree[root_node.name].bcast(rc, root=0)
            spawn.Disconnect()
            assert len(rc) == 1
            rc = rc[0]
        finally:
            os.chdir(currdir)
    else:
        rc, msg = pyutilib.subprocess.run(
            [executable]+args,
            cwd=working_directory,
            outfile=logfile,
            tee=output_solver_log)
    assert rc == 0, str(msg)

    stop = time.time()
    solve_time = stop - start

    #
    # Parse the SOL files for the subproblems local to
    # this worker and load the results
    #
    worker_results = {}
    with ReaderFactory(ResultsFormat.sol) as reader:
        for object_name in filedata:
            start = time.time()
            nl_filename, symbol_map = filedata[object_name]
            assert nl_filename.endswith(".nl")
            sol_filename = nl_filename[:-2]+"sol"
            results = reader(sol_filename, suffixes=suffixes)
            stop = time.time()
            # tag the results object with the symbol_map
            results._smap = symbol_map
            results.solver.time = solve_time
            results.pyomo_solve_time = (stop - start) + \
                                       solve_time + \
                                       write_time[object_name]
            if str(results.solver.termination_condition) == "infeasible":
                if len(results.solution) > 0:
                    results.solution.clear()
            # TODO: Re-architect ScenarioTreeManagerSolver
            #       to better support this
            worker_results[object_name] = \
                load_function(object_name, results)

    return worker_results

def EXTERNAL_collect_solution(worker, node_name):
    solution = {}
    tmp = {}
    node = worker.scenario_tree.get_node(node_name)
    scenario = node.scenarios[0]
    stage = node.stage
    instance = scenario.instance
    assert instance is not None
    bySymbol = instance._ScenarioTreeSymbolMap.bySymbol
    for id_ in node._variable_ids:
        # TODO
        #cost_variable_name, cost_variable_index = \
        #    stage._cost_variable
        #stage_cost_obj = instance.find_component(cost_variable_name)\
        #                 [cost_variable_index]
        #if not stage_cost_obj.is_expression_type():
        #    solution[ComponentUID(stage_cost_obj,cuid_buffer=tmp)] = \
        #        (stage_cost_obj.value, stage_cost_obj.stale)
        for variable_id in node._variable_ids:
            var = bySymbol[variable_id]
            if var.is_expression_type():
                continue
            solution[ComponentUID(var, cuid_buffer=tmp)] = \
                (var.value, var.stale)

    return solution

class SchurIpoptSolver(SPSolverShellCommand, PySPConfiguredObject):

    def __init__(self):
        super(SchurIpoptSolver, self).__init__(
            self.register_options())
        self.set_options_to_default()
        self._executable = "schuripopt"

    def set_options_to_default(self):
        self._options = self.register_options()
        self._options._implicit_declaration = True

    @property
    def options(self):
        return self._options

    @property
    def name(self):
        return "schuripopt"

    def _launch_solver(self,
                       manager,
                       output_directory,
                       logfile,
                       ignore_bundles=False,
                       output_solver_log=False,
                       verbose=False,
                       io_options=None):

        if not os.path.exists(output_directory):
            os.makedirs(output_directory)
        problem_list_filename = os.path.join(output_directory,
                                             "PySP_Subproblems.txt")

        scenario_tree = manager.scenario_tree

        subproblem_type = _write_problem_list_file(
            scenario_tree,
            problem_list_filename,
            ignore_bundles=ignore_bundles)

        assert subproblem_type in ('scenarios','bundles')

        options_filename = os.path.join(output_directory,
                                        "schuripopt.opt")
        # just in case output_directory is not a tmpdir, make sure
        # we don't silently overwrite someone's options file
        assert not os.path.exists(options_filename)
        command_line_options = []
        with open(options_filename, "w") as f:
            for key, val in self.options.items():
                key = key.strip()
                if key.startswith("OF_"):
                    if key == "OF_output_file":
                        raise ValueError(
                            "Use of the 'output_file' option "
                            "is disallowed. Use the logfile "
                            "keyword instead.")
                    f.write(key[3:]+" "+str(val)+"\n")
                else:
                    command_line_options.append((key,val))

        if verbose:
            print("Schuripopt solver problem list file: %s"
                  % (problem_list_filename))
            print("Schuripopt solver options file: %s"
                  % (options_filename))
            print("Schuripopt solver problem type: %s"
                  % (subproblem_type))
            print("Sending solver invocation request to "
                  "workers")

        try:
            worker_results = manager.invoke_function(
                "EXTERNAL_invoke_solve",
                thisfile,
                invocation_type=InvocationType.Single,
                function_args=(output_directory,
                               subproblem_type,
                               logfile,
                               problem_list_filename,
                               self.executable,
                               output_solver_log,
                               io_options,
                               command_line_options,
                               options_filename))
        finally:
            if isinstance(manager,
                          (ScenarioTreeManagerClientPyro,
                           ScenarioTreeManagerSolverClientPyro)) and \
                output_solver_log:
                # If this is Pyro the best we can do is
                # dump the log file to the screen after the
                # solve completes because subprocesses are
                # spawned on the Pyro workers.
                if os.path.exists(logfile):
                    with io.open(logfile) as f:
                        print(f.read())

        results = ScenarioTreeSolveResults(subproblem_type)
        for worker_name in worker_results:
            worker_result = worker_results[worker_name]
            for object_name in worker_result:
                results.update(worker_result[object_name])

        return results

    def _solve_impl(self,
                    sp,
                    output_solver_log=False,
                    verbose=False,
                    logfile=None,
                    **kwds):
        """
        Solve a stochastic program with the SchurIpopt solver.

        See the 'solve' method on the base class for
        additional keyword documentation.

        Args:
            sp: The stochastic program to solve.
            output_solver_log (bool): Stream the solver
                output during the solve.
            logfile: The name of the logfile to save the
                solver output into.
            verbose: Report verbose status information to
                aid debugging.
            **kwds: Passed to the DDSIP file writer as I/O
              options (e.g., symbolic_solver_labels=True).

        Returns: A results object with information about the solution.
        """

        #
        # Setup the SchurIpopt working directory
        #
        problem_list_filename = "PySP_Subproblems.txt"
        working_directory = self._create_tempdir("workdir",
                                                 dir=os.getcwd())

        if logfile is None:
            logfile = os.path.join(working_directory,
                                   "schuripopt.log")
            self._add_tempfile("logfile", logfile)
        else:
            self._files["logfile"] = logfile

        if verbose:
            print("Schuripopt solver working directory: %s"
                  % (working_directory))
            print("Schuripopt solver logfile: %s"
                  % (logfile))

        #
        # Launch SchurIpopt from the worker processes
        # (assumed to be launched together using mpirun)
        #
        status = self._launch_solver(
            sp,
            working_directory,
            logfile=logfile,
            output_solver_log=output_solver_log,
            verbose=verbose,
            io_options=kwds)

        objective = 0.0
        solver_status = set()
        solver_message = set()
        termination_condition = set()
        solution_status = set()
        if status.solve_type == "bundles":
            assert sp.scenario_tree.contains_bundles()
            assert len(status.objective) == \
                len(sp.scenario_tree.bundles)
            for bundle in sp.scenario_tree.bundles:
                if objective is not None:
                    if isinstance(status.objective[bundle.name], UndefinedData):
                        objective = None
                    else:
                        objective += bundle.probability * \
                                     status.objective[bundle.name]
                solver_status.add(status.solver_status[bundle.name])
                solver_message.add(status.solver_message[bundle.name])
                termination_condition.add(status.termination_condition[bundle.name])
                if isinstance(status.solution_status[bundle.name], UndefinedData):
                    solution_status.add(None)
                else:
                    solution_status.add(status.solution_status[bundle.name])
        else:
            assert status.solve_type == "scenarios"
            assert len(status.objective) == \
                len(sp.scenario_tree.scenarios)
            for scenario in sp.scenario_tree.scenarios:
                if objective is not None:
                    if isinstance(status.objective[scenario.name], UndefinedData):
                        objective = None
                    else:
                        objective += scenario.probability * \
                                     status.objective[scenario.name]
                solver_status.add(status.solver_status[scenario.name])
                solver_message.add(status.solver_message[scenario.name])
                termination_condition.add(status.termination_condition[scenario.name])
                if isinstance(status.solution_status[scenario.name], UndefinedData):
                    solution_status.add(None)
                else:
                    solution_status.add(status.solution_status[scenario.name])

        assert len(solver_status) == 1
        assert len(solver_message) == 1
        assert len(termination_condition) == 1
        assert len(solution_status) == 1

        results = SPSolverResults()
        results.objective = None
        results.bound = None
        results.status = solution_status.pop()
        results.solver.status = solver_status.pop()
        results.solver.termination_condition = termination_condition.pop()
        results.solver.message = solver_message.pop()
        results.solver.time = max(status.solve_time.values())
        results.solver.pyomo_time = \
            max(status.pyomo_solve_time.values())

        results.xhat = None
        if str(results.solver.status) == "ok" and \
           str(results.solver.termination_condition) == "optimal":
            results.objective = objective
            xhat = results.xhat = {}
            for stage in sp.scenario_tree.stages[:-1]:
                for node in stage.nodes:
                    worker_name = sp.get_worker_for_scenario(
                        node.scenarios[0].name)
                    node_solution = sp.invoke_function_on_worker(
                        worker_name,
                        "EXTERNAL_collect_solution",
                        thisfile,
                        invocation_type=InvocationType.Single,
                        function_args=(node.name,))
                    node_xhat = xhat[node.name] = {}
                    for id_ in node_solution:
                        node_xhat[repr(id_)] = node_solution[id_][0]

        return results

def runschuripopt_register_options(options=None):
    if options is None:
        options = PySPConfigBlock()
    safe_register_common_option(options,
                               "verbose")
    safe_register_common_option(options,
                               "disable_gc")
    safe_register_common_option(options,
                               "profile")
    safe_register_common_option(options,
                               "traceback")
    safe_register_common_option(options,
                               "output_scenario_tree_solution")
    safe_register_common_option(options,
                                "output_solver_log")
    safe_register_common_option(options,
                                "keep_solver_files")
    safe_register_common_option(options,
                                "symbolic_solver_labels")
    ScenarioTreeManagerFactory.register_options(options)
    SchurIpoptSolver.register_options(options)
    # used to populate the implicit SchurIpopt options
    safe_register_unique_option(
        options,
        "solver_options",
        PySPConfigValue(
            (),
            domain=_domain_tuple_of_str_or_dict,
            description=(
                "Solver options to pass to SchurIpopt "
                "(e.g., relax_integrality=yes). This "
                "option can be used multiple times "
                "from the command line to specify more "
                "than one SchurIpopt option."
            ),
            doc=None,
            visibility=0),
        ap_kwds={'action': 'append'},
        ap_group=_schuripopt_group_label)

    return options

def runschuripopt(options):
    """
    Construct a senario tree manager and solve it
    with the SD solver.
    """
    start_time = time.time()
    with ScenarioTreeManagerFactory(options) as manager:
        manager.initialize()
        print("")
        print("Running SchurIpopt solver for stochastic "
              "programming problems")
        schuripopt = SchurIpoptSolver()
        # add the implicit schuripopt options
        solver_options = options.solver_options
        if len(solver_options) > 0:
            if type(solver_options) is tuple:
                for name_val in solver_options:
                    assert "=" in name_val
                    name, val = name_val.split("=")
                    schuripopt.options[name.strip()] = val.strip()
            else:
                for key, val in solver_options.items():
                    schuripopt.options[key] = val
        results = schuripopt.solve(
            manager,
            output_solver_log=options.output_solver_log,
            keep_solver_files=options.keep_solver_files,
            symbolic_solver_labels=options.symbolic_solver_labels)
        xhat = results.xhat
        del results.xhat
        print("")
        print(results)

        if options.output_scenario_tree_solution:
            print("Final solution (scenario tree format):")
            manager.scenario_tree.snapshotSolutionFromScenarios()
            manager.scenario_tree.pprintSolution()

    print("")
    print("Total execution time=%.2f seconds"
          % (time.time() - start_time))

    return 0

#
# the main driver routine
#

def main(args=None):
    #
    # Top-level command that executes everything
    #

    #
    # Import plugins
    #
    import pyomo.environ

    #
    # Parse command-line options.
    #
    try:
        options = parse_command_line(
            args,
            runschuripopt_register_options,
            prog='runschuripopt',
            description=(
"""Optimize a stochastic program using the SchurIpopt solver."""
            ))

    except SystemExit as _exc:
        # the parser throws a system exit if "-h" is specified
        # - catch it to exit gracefully.
        return _exc.code

    return launch_command(runschuripopt,
                          options,
                          error_label="runschuripopt: ",
                          disable_gc=options.disable_gc,
                          profile_count=options.profile,
                          traceback=options.traceback)

SPSolverFactory.register_solver("schuripopt", SchurIpoptSolver)

if __name__ == "__main__":
    sys.exit(main())



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

import sys
import fnmatch
import os
import subprocess
from os.path import abspath, dirname, join, basename

try:
    from subprocess import check_output as _run_cmd
except:
    # python 2.6
    from subprocess import check_call as _run_cmd

import pyutilib.th as unittest

from pyomo.pysp.tests.examples.ef_checker import main as validate_ef_main

# Global test configuration options
_test_name_wildcard_include = ["*"]
_test_name_wildcard_exclude = [""]
_disable_stdout_test = True
_json_exact_comparison = True
_yaml_exact_comparison = True
_diff_tolerance = 1e-5

#
# Get the directory where this script is defined, and where the baseline
# files are located.
#

thisDir = dirname(abspath(__file__))
baselineDir = join(thisDir,"baselines")
pysp_examples_dir = join(dirname(dirname(dirname(dirname(thisDir)))),"examples","pysp")
pyomo_bin_dir = join(dirname(dirname(dirname(dirname(dirname(dirname(thisDir)))))),"bin")

farmer_examples_dir = join(pysp_examples_dir,"farmer")
farmer_model_dir = join(farmer_examples_dir,"models")
farmer_concrete_model_dir = join(farmer_examples_dir,"concrete")
farmer_expr_model_dir = join(farmer_examples_dir,"expr_models")
farmer_max_model_dir = join(farmer_examples_dir,"maxmodels")
farmer_expr_max_model_dir = join(farmer_examples_dir,"expr_maxmodels")
farmer_data_dir = join(farmer_examples_dir,"scenariodata")
farmer_trivialbundlesdata_dir = join(farmer_examples_dir,"scenariodataWithTrivialBundles")
farmer_other_dir = join(thisDir, "farmer_files")

forestry_examples_dir = join(pysp_examples_dir,"forestry")
forestry_model_dir = join(forestry_examples_dir,"models-nb-yr")
forestry_expr_model_dir = join(forestry_examples_dir,"expr-models-nb-yr")
forestry_data_dir = join(forestry_examples_dir,"18scenarios")
forestry_unequal_data_dir = join(forestry_examples_dir,"unequalProbs")

hydro_examples_dir = join(pysp_examples_dir,"hydro")
hydro_model_dir = join(hydro_examples_dir,"models")
hydro_data_dir = join(hydro_examples_dir,"scenariodata")
hydro_nodedata_dir = join(hydro_examples_dir,"nodedata")

testing_solvers = {}
testing_solvers['cplex','lp'] = False
testing_solvers['cplex','nl'] = False
testing_solvers['ipopt','nl'] = False
testing_solvers['cplex','python'] = False

class EFTester(object):

    baseline_group = None
    model_directory = None
    instance_directory = None
    solver_name = None
    solver_io = None
    base_command_options = ""

    @staticmethod
    def _setUpClass(cls):
        global testing_solvers
        from pyomo.solvers.tests.solvers import test_solver_cases
        for _solver, _io in test_solver_cases():
            if (_solver, _io) in testing_solvers and \
                test_solver_cases(_solver, _io).available:
                testing_solvers[_solver, _io] = True

    def setUp(self):
        assert self.model_directory is not None
        assert self.instance_directory is not None
        assert (self.solver_name,self.solver_io) in testing_solvers
        assert self.base_command_options is not None

    @staticmethod
    def safe_delete(filename):
        try:
            os.remove(filename)
        except OSError:
            pass

    def get_cmd_base(self):
        cmd = ''
        cmd += 'cd '+thisDir+'; '
        cmd += "runef --solve"
        cmd += " --solver="+self.solver_name
        cmd += " --solver-io="+self.solver_io
        cmd += " --traceback "
        cmd += self.base_command_options
        return cmd

    @unittest.nottest
    def _baseline_test(self,
                       options_string="",
                       validation_options_string="",
                       cleanup_func=None,
                       rename_func=None,
                       check_baseline_func=None):

        global _test_name_wildcard_include
        global _test_name_wildcard_exclude
        class_name, test_name = self.id().split('.')[-2:]
        if all(not fnmatch.fnmatch(class_name+'.'+test_name,inc_wildcard) \
               for inc_wildcard in _test_name_wildcard_include):
            self.skipTest("Test %s.%s does not match any include wildcard in '%s'" \
                          % (class_name, test_name, _test_name_wildcard_include))
        if any(fnmatch.fnmatch(class_name+'.'+test_name,exc_wildcard) \
               for exc_wildcard in _test_name_wildcard_exclude):
            self.skipTest("Test %s.%s matches at least one exclude wildcard in '%s'" \
                          % (class_name, test_name, _test_name_wildcard_exclude))
        if not testing_solvers[self.solver_name,self.solver_io]:
            self.skipTest("Solver (%s,%s) not available"
                          % (self.solver_name, self.solver_io))
        prefix = class_name+"."+test_name
        argstring = self.get_cmd_base()+" "\
                    "--model-directory="+self.model_directory+" "\
                    "--instance-directory="+self.instance_directory+" "\
                    "--solution-writer=pyomo.pysp.plugins.jsonsolutionwriter "\
                    +options_string+" "\
                    "&> "+join(thisDir,prefix+".out")
        print("Testing command("+basename(prefix)+"): " + argstring)
        self.safe_delete(join(thisDir,prefix+".out"))
        self.safe_delete(join(thisDir,"ef_solution.json"))
        self.safe_delete(join(thisDir,prefix+".ef_solution.json.out"))
        if cleanup_func is not None:
            cleanup_func(self, class_name, test_name)
        _run_cmd(argstring, shell=True)
        self.assertTrue(os.path.exists(join(thisDir,"ef_solution.json")))
        os.rename(join(thisDir,"ef_solution.json"),
                  join(thisDir,prefix+".ef_solution.json.out"))
        if rename_func is not None:
            rename_func(self, class_name, test_name)

        validate_ef_main([join(thisDir,prefix+".ef_solution.json.out"),
                          '-t',repr(_diff_tolerance)]\
                         +validation_options_string.split())

        if self.baseline_group is not None:
            group_prefix = self.baseline_group+"."+test_name
            # Disable automatic deletion of the ef_solution output
            # file on passing test just in case the optional
            # check_baseline_func wants to look at it.
            self.assertMatchesJsonBaseline(
                join(thisDir,prefix+".ef_solution.json.out"),
                join(baselineDir,group_prefix+".ef_solution.json.baseline.gz"),
                tolerance=_diff_tolerance,
                delete=False,
                exact=_json_exact_comparison)

        if check_baseline_func is not None:
            assert self.baseline_group is not None
            check_baseline_func(self, class_name, test_name)
        else:
            # Now we can safely delete this file because the test has
            # passed if we are here
            self.safe_delete(join(thisDir,prefix+".ef_solution.json.out"))
        self.safe_delete(join(thisDir,prefix+".out"))

    def test1(self):
        self._baseline_test()

    def test2(self):
        self._baseline_test(options_string="--generate-weighted-cvar")

    def test3(self):
        self._baseline_test(options_string="--generate-weighted-cvar "
                                           "--cvar-weight=0 "
                                           "--risk-alpha=0.1")

    def test4(self):
        self._baseline_test(options_string="--generate-weighted-cvar "
                                           "--cvar-weight=0.5")

    def test5(self):
        self._baseline_test(options_string="--generate-weighted-cvar "
                                           "--risk-alpha=0.1")

@unittest.category('expensive')
class TestEFFarmerCPLEXNL(EFTester,unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        EFTester._setUpClass(cls)
        cls.baseline_group = "TestEFFarmer"
        cls.model_directory = join(farmer_concrete_model_dir,'ReferenceModel.py')
        cls.instance_directory = join(farmer_data_dir,'ScenarioStructure.dat')
        cls.solver_name = 'cplex'
        cls.solver_io = 'nl'

@unittest.category('expensive')
class TestEFFarmerCPLEXLP(EFTester,unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        EFTester._setUpClass(cls)
        cls.baseline_group = "TestEFFarmer"
        cls.model_directory = farmer_model_dir
        cls.instance_directory = farmer_data_dir
        cls.solver_name = 'cplex'
        cls.solver_io = 'lp'

@unittest.category('expensive')
class TestEFFarmerCPLEXPYTHON(EFTester,unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        EFTester._setUpClass(cls)
        cls.baseline_group = "TestEFFarmer"
        cls.model_directory = farmer_model_dir
        cls.instance_directory = farmer_data_dir
        cls.solver_name = 'cplex'
        cls.solver_io = 'python'

@unittest.category('expensive')
class TestEFFarmerExpression(EFTester,unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        EFTester._setUpClass(cls)
        cls.baseline_group = "TestEFFarmer"
        cls.model_directory = farmer_expr_model_dir
        cls.instance_directory = farmer_data_dir
        cls.solver_name = 'cplex'
        cls.solver_io = 'lp'

@unittest.category('expensive')
class TestEFFarmerMax(EFTester,unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        EFTester._setUpClass(cls)
        cls.baseline_group = "TestEFFarmerMax"
        cls.model_directory = farmer_max_model_dir
        cls.instance_directory = farmer_data_dir
        cls.solver_name = 'cplex'
        cls.solver_io = 'lp'

@unittest.category('expensive')
class TestEFFarmerMaxExpression(EFTester,unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        EFTester._setUpClass(cls)
        cls.baseline_group = "TestEFFarmerMax"
        cls.model_directory = farmer_expr_max_model_dir
        cls.instance_directory = farmer_data_dir
        cls.solver_name = 'cplex'
        cls.solver_io = 'lp'






@unittest.category('expensive')
class TestEFForestryNoBaseline(EFTester,unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        EFTester._setUpClass(cls)
        cls.baseline_group = None
        cls.model_directory = forestry_model_dir
        cls.instance_directory = forestry_data_dir
        cls.solver_name = 'cplex'
        cls.solver_io = 'lp'
        cls.base_command_options = "--mipgap=0.0125"

@unittest.category('expensive')
class TestEFForestryExpressionNoBaseline(EFTester,unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        EFTester._setUpClass(cls)
        cls.baseline_group = None
        cls.model_directory = forestry_expr_model_dir
        cls.instance_directory = forestry_data_dir
        cls.solver_name = 'cplex'
        cls.solver_io = 'lp'
        cls.base_command_options = "--mipgap=0.0125"

@unittest.category('expensive')
class TestEFForestryUnequalProbsNoBaseline(EFTester,unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        EFTester._setUpClass(cls)
        cls.baseline_group = None
        cls.model_directory = forestry_model_dir
        cls.instance_directory = forestry_unequal_data_dir
        cls.solver_name = 'cplex'
        cls.solver_io = 'lp'
        cls.base_command_options = "--mipgap=0.0125"

@unittest.category('expensive')
class TestEFForestryExpressionUnequalProbsNoBaseline(EFTester,unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        EFTester._setUpClass(cls)
        cls.baseline_group = None
        cls.model_directory = forestry_expr_model_dir
        cls.instance_directory = forestry_unequal_data_dir
        cls.solver_name = 'cplex'
        cls.solver_io = 'lp'
        cls.base_command_options = "--mipgap=0.0125"



@unittest.category('expensive')
class TestEFHydroNoBaseline(EFTester,unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        EFTester._setUpClass(cls)
        cls.baseline_group = None
        cls.model_directory = hydro_model_dir
        cls.instance_directory = hydro_data_dir
        cls.solver_name = 'cplex'
        cls.solver_io = 'lp'

@unittest.category('expensive')
class TestEFHydroNodeDataNoBaseline(EFTester,unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        EFTester._setUpClass(cls)
        cls.baseline_group = None
        cls.model_directory = hydro_model_dir
        cls.instance_directory = hydro_nodedata_dir
        cls.solver_name = 'cplex'
        cls.solver_io = 'lp'


if __name__ == "__main__":

    _disable_stdout_test = False

    import sys
    if '--include' in sys.argv:
        _test_name_wildcard_include = []
        while '--include' in sys.argv:
            idx = sys.argv.index('--include')
            _test_name_wildcard_include.append(sys.argv[idx+1])
            sys.argv.remove('--include')
            sys.argv.remove(_test_name_wildcard_include[-1])
    if '--exclude' in sys.argv:
        _test_name_wildcard_exclude = []
        while '--exclude' in sys.argv:
            idx = sys.argv.index('--exclude')
            _test_name_wildcard_exclude.append(sys.argv[idx+1])
            sys.argv.remove('--exclude')
            sys.argv.remove(_test_name_wildcard_exclude[-1])
    if '--disable-stdout-test' in sys.argv:
        sys.argv.remove('--disable-stdout-test')
        _disable_stdout_test = True

    print("Including all tests matching wildcard: '%s'" % _test_name_wildcard_include)
    print("Excluding all tests matching wildcard: '%s'" % _test_name_wildcard_exclude)

    tester = unittest.main(exit=False)
    if len(tester.result.failures) or len(tester.result.skipped) or len(tester.result.errors):
        with open('UnitTestNoPass.txt','w') as f:
            f.write("Failures:\n")
            for res in tester.result.failures:
                f.write('.'.join(res[0].id().split('.')[-2:])+' ')
            f.write("\n\nSkipped:\n")
            for res in tester.result.skipped:
                f.write('.'.join(res[0].id().split('.')[-2:])+' ')
            f.write("\n\nErrors:\n")
            for res in tester.result.errors:
                f.write('.'.join(res[0].id().split('.')[-2:])+' ')



# the matpolotlib stuff is to avoid $DISPLAY errors on Travis (DLW Oct 2018)
try:
    import matplotlib
    matplotlib.use('Agg')
except:
    pass
from pyomo.common.dependencies import (
    numpy as np, numpy_available,
    pandas as pd, pandas_available,
    scipy, scipy_available,
)
imports_present = numpy_available & pandas_available & scipy_available

uuid_available = True
try:
    import uuid
except:
    uuid_available = False

import pyutilib.th as unittest
import os
import sys

import pyomo.contrib.parmest.parmest as parmest
import pyomo.contrib.parmest.scenariocreator as sc
import pyomo.contrib.parmest.graphics as graphics
import pyomo.contrib.parmest as parmestbase
import pyomo.environ as pyo
import pyomo.contrib.parmest.examples.semibatch.scencreate as sbc

from pyomo.opt import SolverFactory
ipopt_available = SolverFactory('ipopt').available()

testdir = os.path.dirname(os.path.abspath(__file__))


@unittest.skipIf(not imports_present, "Cannot test parmest: required dependencies are missing")
@unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
class pamest_Scenario_creator_reactor_design(unittest.TestCase):
    
    def setUp(self):
        from pyomo.contrib.parmest.examples.reactor_design.reactor_design import reactor_design_model
          
        # Data from the design 
        data = pd.DataFrame(data=[[1.05, 10000, 3458.4, 1060.8, 1683.9, 1898.5],
                                  [1.10, 10000, 3535.1, 1064.8, 1613.3, 1893.4],
                                  [1.15, 10000, 3609.1, 1067.8, 1547.5, 1887.8],
                                  [1.20, 10000, 3680.7, 1070.0, 1486.1, 1881.6],
                                  [1.25, 10000, 3750.0, 1071.4, 1428.6, 1875.0],
                                  [1.30, 10000, 3817.1, 1072.2, 1374.6, 1868.0],
                                  [1.35, 10000, 3882.2, 1072.4, 1324.0, 1860.7],
                                  [1.40, 10000, 3945.4, 1072.1, 1276.3, 1853.1],
                                  [1.45, 10000, 4006.7, 1071.3, 1231.4, 1845.3],
                                  [1.50, 10000, 4066.4, 1070.1, 1189.0, 1837.3],
                                  [1.55, 10000, 4124.4, 1068.5, 1148.9, 1829.1],
                                  [1.60, 10000, 4180.9, 1066.5, 1111.0, 1820.8],
                                  [1.65, 10000, 4235.9, 1064.3, 1075.0, 1812.4],
                                  [1.70, 10000, 4289.5, 1061.8, 1040.9, 1803.9],
                                  [1.75, 10000, 4341.8, 1059.0, 1008.5, 1795.3],
                                  [1.80, 10000, 4392.8, 1056.0,  977.7, 1786.7],
                                  [1.85, 10000, 4442.6, 1052.8,  948.4, 1778.1],
                                  [1.90, 10000, 4491.3, 1049.4,  920.5, 1769.4],
                                  [1.95, 10000, 4538.8, 1045.8,  893.9, 1760.8]], 
                          columns=['sv', 'caf', 'ca', 'cb', 'cc', 'cd'])

        theta_names = ['k1', 'k2', 'k3']
        
        def SSE(model, data): 
            expr = (float(data['ca']) - model.ca)**2 + \
                   (float(data['cb']) - model.cb)**2 + \
                   (float(data['cc']) - model.cc)**2 + \
                   (float(data['cd']) - model.cd)**2
            return expr
        
        self.pest = parmest.Estimator(reactor_design_model, data, theta_names, SSE)

    def test_scen_from_exps(self):
        scenmaker = sc.ScenarioCreator(self.pest, "ipopt")
        experimentscens = sc.ScenarioSet("Experiments")
        scenmaker.ScenariosFromExperiments(experimentscens)
        experimentscens.write_csv("delme_exp_csv.csv")
        df = pd.read_csv("delme_exp_csv.csv")
        os.remove("delme_exp_csv.csv")
        # March '20: all reactor_design experiments have the same theta values!
        k1val = df.loc[5].at["k1"] 
        self.assertAlmostEqual(k1val, 5.0/6.0, places=2)
        tval = experimentscens.ScenarioNumber(0).ThetaVals["k1"]
        self.assertAlmostEqual(tval, 5.0/6.0, places=2)

        
    @unittest.skipIf(not uuid_available, "The uuid module is not available")
    def test_no_csv_if_empty(self):
        # low level test of scenario sets
        # verify that nothing is written, but no errors with empty set

        emptyset = sc.ScenarioSet("empty")
        tfile = uuid.uuid4().hex+".csv"
        emptyset.write_csv(tfile)
        self.assertFalse(os.path.exists(tfile),
                         "ScenarioSet wrote csv in spite of empty set")

        


@unittest.skipIf(not imports_present, "Cannot test parmest: required dependencies are missing")
@unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
class  pamest_Scenario_creator_semibatch(unittest.TestCase):
    
    def setUp(self):
        import pyomo.contrib.parmest.examples.semibatch.semibatch as sb
        import json

        # Vars to estimate in parmest
        theta_names = ['k1', 'k2', 'E1', 'E2']

        self.fbase = os.path.join(testdir,"..","examples","semibatch")
        # Data, list of dictionaries
        data = [] 
        for exp_num in range(10):
            fname = "exp"+str(exp_num+1)+".out"
            fullname = os.path.join(self.fbase, fname)
            with open(fullname,'r') as infile:
                d = json.load(infile)
                data.append(d)

        # Note, the model already includes a 'SecondStageCost' expression 
        # for the sum of squared error that will be used in parameter estimation

        self.pest = parmest.Estimator(sb.generate_model, data, theta_names)
        

    def test_semibatch_bootstrap(self):

        scenmaker = sc.ScenarioCreator(self.pest, "ipopt")
        bootscens = sc.ScenarioSet("Bootstrap")
        numtomake = 2
        scenmaker.ScenariosFromBoostrap(bootscens, numtomake, seed=1134)
        tval = bootscens.ScenarioNumber(0).ThetaVals["k1"]
        self.assertAlmostEqual(tval, 20.64, places=1)

    def test_semibatch_example(self):
        # this is referenced in the documentation so at least look for smoke
        sbc.main(self.fbase)
        
if __name__ == '__main__':
    unittest.main()



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________
#
# Test NEOS solver interface
#

import os
from os.path import abspath, dirname, join
currdir = dirname(abspath(__file__))

import pyutilib.th as unittest

from pyomo.common.dependencies import yaml, yaml_available, yaml_load_args
import pyomo.scripting.pyomo_command as main
from pyomo.scripting.util import cleanup
from pyomo.neos.kestrel import kestrelAMPL

import pyomo.environ as pyo

from six import iteritems

neos_available = False
try:
    # Attempt a connection to NEOS.  Any failure will result in skipping
    # these tests
    if kestrelAMPL().neos is not None:
        neos_available = True
except:
    pass


#
# Because the Kestrel tests require connections to the NEOS server, and
# that can take quite a while (5-20+ seconds), we will only run these
# tests as part of the nightly suite (i.e., by the CI system as part of
# PR / master tests)
#
@unittest.category('nightly')
@unittest.skipIf(not neos_available, "Cannot make connection to NEOS server")
class TestKestrel(unittest.TestCase):

    @unittest.skipIf(not yaml_available, "YAML is not available")
    def test_pyomo_command(self):
        results = os.path.join(currdir, 'result.yml')
        args = [
            os.path.join(currdir,'t1.py'),
            '--solver-manager=neos',
            '--solver=cbc',
            '--symbolic-solver-labels',
            '--save-results=%s' % results,
            '-c',
            ]
        try:
            output = main.run(args)
            self.assertEqual(output.errorcode, 0)

            with open(results) as FILE:
                data = yaml.load(FILE, **yaml_load_args)
            self.assertEqual(
                data['Solver'][0]['Status'], 'ok')
            self.assertAlmostEqual(
                data['Solution'][1]['Status'], 'optimal')
            self.assertAlmostEqual(
                data['Solution'][1]['Objective']['o']['Value'], 1)
            self.assertAlmostEqual(
                data['Solution'][1]['Variable']['x']['Value'], 0.5)
        finally:
            cleanup()
            os.remove(results)

    def test_kestrel_plugin(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(0,1), initialize=0)
        m.c = pyo.Constraint(expr=m.x <= 0.5)
        m.obj = pyo.Objective(expr=2*m.x, sense=pyo.maximize)

        solver_manager = pyo.SolverManagerFactory('neos')
        results = solver_manager.solve(m, opt='cbc')

        self.assertEqual(results.solver[0].status, pyo.SolverStatus.ok)
        self.assertAlmostEqual(pyo.value(m.x), 0.5)


if __name__ == "__main__":
    unittest.main()



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

import sys
import shelve
import json
import os

# TODO:
# - check what weights are for fixed variables (should be zero)
# - check that fix queue from previous iter is always pushed

# don't use this anywhere but this file, it is ugly
def _update_exception_message(msg, e):
    # assert without any args means a coding error
    # in this file, otherwise we want to stack the assertion
    # messages in order to create more informative output
    if len(e.args):
        e.args = (msg+"\n"+e.args[0],)+e.args[1:]
        try:
            e.message = msg + "\n" + e.message
        except:
            # Not python 2
            pass

def assert_value_equals(x,y):
    if x != y:
        raise AssertionError("%s != %s\n" % (x,y))

def assert_float_gt(ub,lb,tol):
    if (ub is None) or (lb is None):
        raise AssertionError("%s !>= %s\n" % (ub,lb))
    if lb-ub > tol:
        raise AssertionError("%.17g !>= %.17g\n" % (ub,lb))

def assert_float_equals(x,y,tol):
    if (x is None) or (y is None):
        raise AssertionError("%s != %s\n" % (x,y))
    if abs(x-y) > tol:
        raise AssertionError("%.17g != %.17g\n" % (x,y))

def validate_probabilities(options, scenario_tree):

    for scenario_name, scenario in scenario_tree['scenarios'].items():
        scenario_probability = scenario['probability']
        test_probability = 1.0
        node_variables = set()
        # check that the probabilities on the scenario tree make sense
        for node_name in scenario['nodes']:
            test_probability *= \
                scenario_tree['nodes'][node_name]['conditional probability']
        try:
            assert_float_equals(scenario_probability,
                                test_probability,
                                options.absolute_tolerance)
        except AssertionError as e:
            msg = ""
            msg += ("Problem validating scenario tree probabilities "
                    "for scenario %s. Product of node conditional "
                    "probabilities does not match scenario probability.\n"
                    % (scenario_name))
            msg += ("Scenario Probability: %r\n" % (scenario_probability))
            msg += ("Node Conditional Product: %r\n" % (test_probability))
            for node_name in scenario['nodes']:
                msg += ("\t%s: %r\n"
                        % (node_name,
                           scenario_tree['nodes'][node_name]['conditional probability']))
            _update_exception_message(msg, e)
            raise

def validate_variable_set(options, scenario_tree, scenario_solutions, node_solutions):

    for scenario_name, scenario in scenario_tree['scenarios'].items():
        scenario_variable_set = \
            set(scenario_solutions[scenario_name]['variables'].keys())
        node_variable_set = set()
        for node_name in scenario['nodes']:
            node_variable_set.update(node_solutions[node_name]['variables'].keys())

        try:
            assert node_variable_set == scenario_variable_set
        except AssertionError as e:
            msg = ("Problem validating the set of reported node "
                   "variables against the set of reported scenario "
                   "variables. Sets do not match:\n")
            msg += ("Symmetric Difference: %s\n"
                    % (set.symmetric_difference(node_variable_set,
                                                scenario_variable_set)))
            raise AssertionError(msg)

def validate_leaf_stage_costs(options,
                              scenario_tree,
                              scenario_solutions,
                              node_solutions):

    # check that the stage cost of the last stage for scenario solutions
    # is equal to the expected node cost for the corresponding leaf node
    # solution
    for scenario_name, scenario in scenario_tree['scenarios'].items():
        scenario_sol = scenario_solutions[scenario_name]
        for stage_name, stage_cost in scenario_sol['stage costs'].items():
            found_one = False
            for node_name in scenario['nodes']:
                node = scenario_tree['nodes'][node_name]
                node_sol = node_solutions[node_name]
                if node['stage'] == stage_name:
                    found_one = True
                    if len(node['children']) == 0:
                        try:
                            assert_float_equals(stage_cost,
                                                node_sol['expected cost'],
                                                options.absolute_tolerance)
                        except AssertionError as e:
                            msg = ""
                            msg += ("Problem validating leaf stage cost for "
                                    "scenario %s. Expected node cost in leaf "
                                    "node solution does not match final stage "
                                    "cost in scenario solution.\n"
                                    % (scenario_name))
                            msg += ("Expected cost of node %s: %r\n"
                                    % (node_name, node_sol['expected cost']))
                            msg += ("Scenario cost of stage %s: %r\n"
                                    % (stage_name, stage_cost))
                            _update_exception_message(msg, e)
                            raise

            assert found_one

def validate_scenario_costs(options, scenario_tree, scenario_solutions):
    for scenario_name, scenario in scenario_tree['scenarios'].items():
        assert scenario_name == scenario['name']
        scenario_sol = scenario_solutions[scenario_name]

        try:
            assert scenario_sol['objective'] is not None, "value cannot be None"
            assert scenario_sol['cost'] is not None, "value cannot be None"
            assert scenario_sol['ph weight term'] is not None, "value cannot be None"
            assert scenario_sol['ph proximal term'] is not None, "value cannot be None"
            assert_float_equals(scenario_sol['objective'],
                                scenario_sol['cost']+\
                                scenario_sol['ph weight term']+\
                                scenario_sol['ph proximal term'],
                                options.absolute_tolerance)
        except AssertionError as e:
            msg = ""
            msg += ("Problem validating objective terms for "
                    "scenario %s. Sum of cost and ph terms does not "
                    "match objective\n" % (scenario_name))
            msg += ("Objective: %r\n" % (scenario_sol['objective']))
            msg += ("Cost: %r\n" % (scenario_sol['cost']))
            msg += ("PH Weight Term: %r\n" % (scenario_sol['ph weight term']))
            msg += ("PH Proximal Term: %r\n" % (scenario_sol['ph proximal term']))
            _update_exception_message(msg, e)
            raise

        try:
            assert_float_equals(scenario_sol['cost'],
                                sum(scenario_sol['stage costs'][stage_name] \
                                    for stage_name in scenario_tree['stages']),
                                options.absolute_tolerance)
        except AssertionError as e:
            msg = ""
            msg += ("Problem validating stage costs terms for "
                    "scenario %s. Sum of stage costs does not match "
                    "value of total cost term used in objective function.\n"
                    % (scenario_name))
            msg += ("Cost: %.17g\n" % (scenario_sol['cost']))
            for stage_name in scenario_tree['stages']:
                msg += ("%s Cost: %.17g\n" % (stage_name, scenario_sol['stage costs'][stage_name]))
            _update_exception_message(msg, e)
            raise

def validate_variable_statistics(options,
                                 scenario_tree,
                                 node_solutions,
                                 scenario_solutions):

    for node_name, node in scenario_tree['nodes'].items():
        try:
            assert node_name == node['name']
            node_sol = node_solutions[node_name]
            is_leaf_node = bool(len(node['children']) == 0)

            for scenario_name in node['scenarios']:
                scenario_sol = scenario_solutions[scenario_name]

            variable_wbar = dict.fromkeys(node_sol['variables'],0)
            variable_avg = dict.fromkeys(node_sol['variables'],0)
            variable_min = dict.fromkeys(node_sol['variables'],None)
            variable_max = dict.fromkeys(node_sol['variables'],None)
            variable_fixed_count = dict.fromkeys(node_sol['variables'],0)
            variable_fixed_value = \
                dict((variable_name,dict.fromkeys(node['scenarios'],None)) \
                     for variable_name in node_sol['variables'])

            for variable_name in node_sol['variables']:
                for scenario_name in node['scenarios']:
                    probability = scenario_tree['scenarios'][scenario_name]['probability']
                    scenario_sol = scenario_solutions[scenario_name]
                    scenario_variables_sol = scenario_sol['variables']
                    var_sol = scenario_sol['variables'][variable_name]
                    varval = var_sol['value']
                    variable_avg[variable_name] += probability*varval
                    if var_sol['weight'] is not None:
                        variable_wbar[variable_name] += \
                            probability*var_sol['weight']
                    else:
                        variable_wbar[variable_name] = None

                    if (variable_min[variable_name] is None) or \
                       (varval < variable_min[variable_name]):
                        variable_min[variable_name] = varval
                    if (variable_max[variable_name] is None) or \
                       (varval > variable_max[variable_name]):
                        variable_max[variable_name] = varval
                    variable_fixed_count[variable_name] += int(var_sol['fixed'])
                    if var_sol['fixed']:
                        variable_fixed_value[variable_name][scenario_name] = varval

            for variable_name, variable_sol in node_sol['variables'].items():
                scaled_avg = variable_avg[variable_name]/node['probability']
                # validate fixed count
                try:
                    if variable_sol['fixed']:
                        assert_value_equals(variable_fixed_count[variable_name],
                                            len(node['scenarios']))
                    else:
                        assert_value_equals(variable_fixed_count[variable_name],
                                            0)
                except AssertionError as e:
                    msg = ""
                    msg += ("Problem validating fixed status for variable %s. "
                            "The fixed status of this variable is not consistent "
                            "across all scenarios within this node.\n"
                            % (variable_name))
                    _update_exception_message(msg, e)
                    raise

                if variable_sol['fixed']:
                    scenario_name = None
                    scenario_value = None
                    try:
                        dict_items = list(variable_fixed_value[variable_name].items())
                        for (scen1_name, scen1_val), (scen2_name, scen2_val) in zip(dict_items, dict_items[1:]):
                            scenario_name = scen1_name
                            scenario_value = scen1_val
                            assert_float_equals(scen1_val,
                                                scen2_val,
                                                options.absolute_tolerance)
                        for scen_name, scen_val in dict_items:
                            scenario_name = scen_name
                            scenario_value = scen_val
                            assert_float_equals(scen_val,
                                                scaled_avg,
                                                options.absolute_tolerance)

                    except AssertionError as e:
                        msg = ""
                        msg += ("Problem validating fixed status for variable %s "
                                "in scenario %s. Fixed value does not match that of "
                                "other scenarios in this node or that of node average.\n"
                                % (variable_name, scenario_name))
                        msg += ("Scenario %s fixed value: %.17g\n"
                                % (scenario_name, scenario_value))
                        msg += ("Computed average from scenario solutions: %.17g\n"
                                % (scaled_avg))
                        _update_exception_message(msg, e)
                        raise
                    del scenario_name
                    del scenario_value

                # validate node "solution" (average)
                try:
                    assert_float_equals(variable_sol['solution'],
                                        scaled_avg,
                                        options.absolute_tolerance)
                except AssertionError as e:
                    msg = ""
                    msg += ("Problem validating node 'solution' for variable %s\n"
                            % (variable_name))
                    msg += ("Node 'solution' in solution: %.17g\n"
                            % variable_sol['solution'])
                    msg += ("Computed average from scenario solutions: %.17g\n"
                            % (scaled_avg))
                    _update_exception_message(msg, e)
                    raise

                if not is_leaf_node:
                    # validate node average
                    try:
                        assert_float_equals(variable_sol['average'],
                                            scaled_avg,
                                            options.absolute_tolerance)
                    except AssertionError as e:
                        msg = ""
                        msg += ("Problem validating node average for variable %s\n"
                                % (variable_name))
                        msg += ("Node average in solution: %.17g\n"
                                % variable_sol['average'])
                        msg += ("Computed average from scenario solutions: %.17g\n"
                                % (scaled_avg))
                        _update_exception_message(msg, e)
                        raise

                    # validate node max
                    try:
                        assert_float_equals(variable_sol['maximum'],
                                            variable_max[variable_name],
                                            options.absolute_tolerance)
                    except AssertionError as e:
                        msg = ""
                        msg += ("Problem validating node maximum for variable %s\n"
                                % (variable_name))
                        msg += ("Node maximum in solution: %.17g\n"
                                % variable_sol['maximum'])
                        msg += ("Computed maximum from scenario solutions: %.17g\n"
                                % (variable_max[variable_name]))
                        _update_exception_message(msg, e)
                        raise

                    # validate node min
                    try:
                        assert_float_equals(variable_sol['minimum'],
                                            variable_min[variable_name],
                                            options.absolute_tolerance)
                    except AssertionError as e:
                        msg = ""
                        msg += ("Problem validating node minimum for variable %s\n"
                                % (variable_name))
                        msg += ("Node minimum in solution: %r\n"
                                % variable_sol['minimum'])
                        msg += ("Computed minimum from scenario solutions: %r\n"
                                % (variable_min[variable_name]))
                        _update_exception_message(msg, e)
                        raise

                    # validate wbar
                    try:
                        assert_float_equals(variable_sol['wbar'],
                                            variable_wbar[variable_name],
                                            options.absolute_tolerance)
                    except AssertionError as e:
                        msg = ""
                        msg += ("Problem validating node wbar for variable %s\n"
                                % (variable_name))
                        msg += ("Node wbar in solution: %r\n"
                                % variable_sol['wbar'])
                        msg += ("Computed wbar from scenario solutions: %r\n"
                                % (variable_wbar[variable_name]))
                        _update_exception_message(msg, e)
                        raise

        except AssertionError as e:
            msg = ("Problem validating node %s" % (node_name))
            _update_exception_message(msg, e)
            raise

def validate_ph_objective_parameters(options,
                                     scenario_tree,
                                     scenario_solutions_previous,
                                     node_solutions_previous,
                                     scenario_solutions_current,
                                     node_solutions_current):

    # check that sum of weigth and rho and xbar from prev iter
    # with current variable sols compute the current sol's
    # ph objective terms
    computed_proximal_term = dict.fromkeys(scenario_tree['scenarios'],0.0)
    computed_weight_term = dict.fromkeys(scenario_tree['scenarios'],0.0)
    for node_name, node in scenario_tree['nodes'].items():
        assert node_name == node['name']
        prev_node_vars_sol = node_solutions_previous[node_name]['variables']
        curr_node_vars_sol = node_solutions_current[node_name]['variables']
        is_leaf_node = bool(len(node['children']) == 0)
        if not is_leaf_node:
            for variable_name, prev_node_variable_sol in prev_node_vars_sol.items():
                curr_node_variable_sol = curr_node_vars_sol[variable_name]
                derived = prev_node_variable_sol['derived']
                assert derived == curr_node_variable_sol['derived']
                if derived:
                    continue
                prev_var_xbar = prev_node_variable_sol['xbar']
                for scenario_name in node['scenarios']:
                    prev_scenario_sol = scenario_solutions_previous[scenario_name]
                    prev_var_scenario_sol = \
                        prev_scenario_sol['variables'][variable_name]
                    curr_scenario_sol = scenario_solutions_current[scenario_name]
                    curr_var_scenario_sol = \
                        curr_scenario_sol['variables'][variable_name]
                    computed_proximal_term[scenario_name] += \
                            prev_var_scenario_sol['rho']*0.5*\
                            (curr_var_scenario_sol['value'] - \
                             prev_var_xbar)**2
                    computed_weight_term[scenario_name] += \
                            prev_var_scenario_sol['weight']*\
                            curr_var_scenario_sol['value']

    for scenario_name, scenario in scenario_tree['scenarios'].items():
        curr_scenario_sol = scenario_solutions_current[scenario_name]
        try:
            if options.proximal_is_lb:
                assert_float_gt(computed_proximal_term[scenario_name],
                                curr_scenario_sol['ph proximal term'],
                                options.absolute_tolerance)
            elif options.proximal_is_ub:
                assert_float_gt(curr_scenario_sol['ph proximal term'],
                                computed_proximal_term[scenario_name],
                                options.absolute_tolerance)
            elif not options.disable_proximal_check:
                assert_float_equals(computed_proximal_term[scenario_name],
                                    curr_scenario_sol['ph proximal term'],
                                    options.absolute_tolerance)
        except AssertionError as e:
            msg = ""
            msg += ("Reported Proximal Term: %.17g\n"
                    % (curr_scenario_sol['ph proximal term']))
            msg += ("Computed Proximal Term: %.17g\n"
                    % (computed_proximal_term[scenario_name]))
            msg += ("Problem validating ph proximal term for "
                    "scenario %s. Value reported in current solution "
                    "does not %s value computed using ph parameters "
                    "reported in previous iteration with variable solution "
                    "reported in current iteration.\n"
                    % (scenario_name,
                       "correctly bound" if (options.proximal_is_lb or \
                                             options.proximal_is_ub)
                                         else
                       "match"))
            _update_exception_message(msg, e)
            raise
        try:
            assert_float_equals(computed_weight_term[scenario_name],
                                curr_scenario_sol['ph weight term'],
                                options.absolute_tolerance)
        except AssertionError as e:
            msg = ""
            msg += ("Reported Weight Term: %.17g\n"
                    % (curr_scenario_sol['ph weight term']))
            msg += ("Computed Weight Term: %.17g\n"
                    % (computed_weight_term[scenario_name]))
            msg += ("Problem validating ph weight term for "
                    "scenario %s. Value reported in current solution "
                    "does not match value computed using ph parameters "
                    "reported in previous iteration with variable solution "
                    "reported in current iteration.\n" % (scenario_name))
            _update_exception_message(msg, e)
            raise

def validate_ph(options):

    history = None
    try:
        with open(options.ph_history_filename) as f:
            history = json.load(f)
    except:
        history = None
        try:
            history = shelve.open(options.ph_history_filename,
                                  flag='r')
        except:
            history = None

    if history is None:
        raise RuntimeError("Unable to open ph history file as JSON "
                           "or python Shelve DB format")

    scenario_tree = history['scenario tree']

    try:
        validate_probabilities(options,
                               scenario_tree)
    except AssertionError as e:
        msg = "FAILURE:\n"
        msg += ("Problem validating probabilities for scenario "
                "tree in ph history from file: %s"
                % (options.ph_history_filename))
        _update_exception_message(msg, e)
        raise

    try:
        iter_keys = history['results keys']
    except KeyError:
        # we are using json format (which loads the entire file
        # anyway)
        iter_keys = list(history.keys())
        iter_keys.remove('scenario tree')
    iterations = sorted(int(k) for k in iter_keys)
    iterations = [str(k) for k in iterations]

    prev_iter_results = None
    for i in iterations:

        iter_results = history[i]

        convergence = iter_results['convergence']
        scenario_solutions = iter_results['scenario solutions']
        node_solutions = iter_results['node solutions']

        try:
            validate_variable_set(options,
                                  scenario_tree,
                                  scenario_solutions,
                                  node_solutions)
        except AssertionError as e:
            msg = "FAILURE:\n"
            msg += ("Problem validating variable set for iteration %s "
                    "of ph history in file: %s"
                    % (i, options.ph_history_filename))
            _update_exception_message(msg, e)
            raise


        try:
            validate_leaf_stage_costs(options,
                                      scenario_tree,
                                      scenario_solutions,
                                      node_solutions)
        except AssertionError as e:
            msg = "FAILURE:\n"
            msg += ("Problem validating leaf-stage costs for iteration %s "
                    "of ph history in file: %s"
                    % (i, options.ph_history_filename))
            _update_exception_message(msg, e)
            raise


        try:
            validate_scenario_costs(options, scenario_tree, scenario_solutions)
        except AssertionError as e:
            msg = "FAILURE:\n"
            msg += ("Problem validating scenario costs for iteration %s "
                    "of ph history in file: %s"
                    % (i, options.ph_history_filename))
            _update_exception_message(msg, e)
            raise


        try:
            validate_scenario_costs(options, scenario_tree, scenario_solutions)
        except AssertionError as e:
            msg = "FAILURE:\n"
            msg += ("Problem validating scenario costs for iteration %s "
                    "of ph history in file: %s"
                    % (i, options.ph_history_filename))
            _update_exception_message(msg, e)
            raise

        try:
            validate_variable_statistics(options,
                                         scenario_tree,
                                         node_solutions,
                                         scenario_solutions)
        except AssertionError as e:
            msg = "FAILURE:\n"
            msg += ("Problem validating variable statistics for iteration %s "
                    "of ph history in file: %s"
                    % (i, options.ph_history_filename))
            _update_exception_message(msg, e)
            raise

        if prev_iter_results is not None:
            try:
                validate_ph_objective_parameters(
                    options,
                    scenario_tree,
                    prev_iter_results['scenario solutions'],
                    prev_iter_results['node solutions'],
                    scenario_solutions,
                    node_solutions)
            except AssertionError as e:
                msg = "FAILURE:\n"
                msg += ("Problem validating ph objective parameters "
                        "for iteration %s of ph history in file: %s"
                        % (i, options.ph_history_filename))
                _update_exception_message(msg, e)
                raise
        prev_iter = i

    return 0

def construct_parser():

    import argparse
    parser = argparse.ArgumentParser(
        description=('Validate a JSON ph history output file that '
                     'was created the phhistoryextension plugin'))
    parser.add_argument('ph_history_filename', metavar='JSON_HISTORY', type=str,
                        action='store', default=None,
                        help=('JSON formatted output file to '
                              'validate'))
    parser.add_argument('-t','--tolerance',dest='absolute_tolerance',
                        action='store', default=1e-6, type=float,
                        help=('absolute tolerance used when comparing '
                              'floating point values (default: %(default)s)'))
    group1 = parser.add_mutually_exclusive_group()
    group1.add_argument('--proximal-term-bounds-below', dest='proximal_is_lb',
                        action='store_true', default=False,
                        help=('relax proximal term check from equality and '
                              'ensure that the solution represents a lower '
                              'bound'))
    group1.add_argument('--proximal-term-bounds-above', dest='proximal_is_ub',
                        action='store_true', default=False,
                        help=('relax proximal term check from equality and '
                              'ensure that the solution represents an upper '
                              'bound'))
    group1.add_argument('--disable-proximal-term-check', dest='disable_proximal_check',
                        action='store_true', default=False,
                        help=('skip validation of proximal term recomputation'))

    return parser

def main(args=None):

    parser = construct_parser()
    options = parser.parse_args(args=args)
    assert os.path.exists(options.ph_history_filename)
    return validate_ph(options)

if __name__ == "__main__":

    import sys
    sys.exit(main(args=sys.argv[1:]))



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

from __future__ import division
import math

#
# Provide a global value that indicates which expression system is being used
#
class Mode(object):
    pyomo5_trees = (3,)
_mode = Mode.pyomo5_trees

#
# Common intrinsic functions
#
from pyomo.core.expr import expr_common as common
#
# Pull symbols from the appropriate expression system
#
from pyomo.core.expr import numvalue as _numvalue
from pyomo.core.expr import boolean_value as _logicalvalue

# Pyomo5
if _mode == Mode.pyomo5_trees:
    from pyomo.core.expr import numeric_expr as _numeric_expr
    from pyomo.core.expr.numeric_expr import *
    from pyomo.core.expr.numeric_expr import (
        _generate_sum_expression,
        _generate_mul_expression,
        _generate_other_expression,
        _generate_intrinsic_function_expression,
    )
    from pyomo.core.expr import logical_expr as _logical_expr
    from pyomo.core.expr.logical_expr import *
    from pyomo.core.expr.logical_expr import (
        _generate_relational_expression,
        _chainedInequality,
    )
    from pyomo.core.expr.template_expr import *
    from pyomo.core.expr import visitor as _visitor
    from pyomo.core.expr.visitor import *
    # FIXME: we shouldn't need circular dependencies between modules
    _visitor.LinearExpression = _numeric_expr.LinearExpression
    _visitor.MonomialTermExpression = _numeric_expr.MonomialTermExpression
    _visitor.NPV_expression_types = _numeric_expr.NPV_expression_types
    _visitor.clone_counter = _numeric_expr.clone_counter

    # Initialize numvalue functions
    _numvalue._generate_sum_expression \
        = _numeric_expr._generate_sum_expression
    _numvalue._generate_mul_expression \
        = _numeric_expr._generate_mul_expression
    _numvalue._generate_other_expression \
        = _numeric_expr._generate_other_expression
    _numvalue._generate_relational_expression \
        = _logical_expr._generate_relational_expression

    # Initialize logicalvalue functions
    _logicalvalue._generate_logical_proposition = _logical_expr._generate_logical_proposition
else:
    raise ValueError("No other expression systems are supported in Pyomo right now.")    #pragma: no cover


def Expr_if(IF=None, THEN=None, ELSE=None):
    """
    Function used to construct a logical conditional expression.
    """
    return Expr_ifExpression(IF_=IF, THEN_=THEN, ELSE_=ELSE)

#
# NOTE: abs() and pow() are not defined here, because they are
# Python operators.
#
def ceil(arg):
    return _generate_intrinsic_function_expression(arg, 'ceil', math.ceil)

def floor(arg):
    return _generate_intrinsic_function_expression(arg, 'floor', math.floor)

# e ** x
def exp(arg):
    return _generate_intrinsic_function_expression(arg, 'exp', math.exp)

def log(arg):
    return _generate_intrinsic_function_expression(arg, 'log', math.log)

def log10(arg):
    return _generate_intrinsic_function_expression(arg, 'log10', math.log10)

# FIXME: this is nominally the same as x ** 0.5, but follows a different
# path and produces a different NL file!
def sqrt(arg):
    return _generate_intrinsic_function_expression(arg, 'sqrt', math.sqrt)
#    return _generate_expression(common._pow, arg, 0.5)


def sin(arg):
    return _generate_intrinsic_function_expression(arg, 'sin', math.sin)

def cos(arg):
    return _generate_intrinsic_function_expression(arg, 'cos', math.cos)

def tan(arg):
    return _generate_intrinsic_function_expression(arg, 'tan', math.tan)

def sinh(arg):
    return _generate_intrinsic_function_expression(arg, 'sinh', math.sinh)

def cosh(arg):
    return _generate_intrinsic_function_expression(arg, 'cosh', math.cosh)

def tanh(arg):
    return _generate_intrinsic_function_expression(arg, 'tanh', math.tanh)


def asin(arg):
    return _generate_intrinsic_function_expression(arg, 'asin', math.asin)

def acos(arg):
    return _generate_intrinsic_function_expression(arg, 'acos', math.acos)

def atan(arg):
    return _generate_intrinsic_function_expression(arg, 'atan', math.atan)

def asinh(arg):
    return _generate_intrinsic_function_expression(arg, 'asinh', math.asinh)

def acosh(arg):
    return _generate_intrinsic_function_expression(arg, 'acosh', math.acosh)

def atanh(arg):
    return _generate_intrinsic_function_expression(arg, 'atanh', math.atanh)



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

import gc         # garbage collection control.
import os
import socket
import sys
import time
import copy
from optparse import OptionParser

import pyutilib.common
import pyutilib.misc
from pyutilib.misc import PauseGC
from pyutilib.pyro import (TaskWorker,
                           TaskWorkerServer,
                           shutdown_pyro_components)

from pyomo.core import *
from pyomo.opt import UndefinedData
from pyomo.common import pyomo_command
from pyomo.common.plugin import ExtensionPoint, SingletonPlugin
from pyomo.opt import (SolverFactory,
                       TerminationCondition,
                       SolutionStatus)
from pyomo.pysp.phextension import IPHSolverServerExtension
from pyomo.pysp.scenariotree.instance_factory import \
    ScenarioTreeInstanceFactory
from pyomo.pysp.phsolverserverutils import (TransmitType,
                                           InvocationType)
from pyomo.pysp.ph import _PHBase
from pyomo.pysp.phutils import (reset_nonconverged_variables,
                                reset_stage_cost_variables,
                                find_active_objective,
                                extract_solve_times)
from pyomo.pysp.util.misc import launch_command

from six import iterkeys, iteritems

class PHPyroWorker(TaskWorker):

    def __init__(self, **kwds):

        # add for purposes of diagnostic output.
        kwds["caller_name"] = "PH Pyro Server"
        kwds["name"] = ("PySPWorker_%d@%s" % (os.getpid(), socket.gethostname()))
        TaskWorker.__init__(self, **kwds)

        self.type = self.WORKERNAME
        self.block = True
        self.timeout = None
        self._phsolverserver_map = {}
        # phsolverserver processes all collect from different queues,
        # so we can collect as many tasks as are available in the queue
        self._bulk_task_collection = True
        self._modules_imported = {}

    def process(self, data):

        data = pyutilib.misc.Bunch(**data)
        result = None
        if data.action == "release":

            del self._phsolverserver_map[name]
            result = True

        elif data.action == "initialize":

            self._phsolverserver_map[data.object_name] = _PHSolverServer(self._modules_imported)
            self._phsolverserver_map[data.object_name].WORKERNAME = self.WORKERNAME
            data.name = data.object_name
            result = self._phsolverserver_map[data.name].process(data)

        elif data.action == "shutdown":

            print("Received shutdown request")
            self.dispatcher.clear_queue(type=self.type)
            self._worker_shutdown = True

        else:

            with PauseGC() as pgc:
                result = self._phsolverserver_map[data.name].process(data)

        return result

class _PHSolverServer(_PHBase):

    def __init__(self, modules_imported, **kwds):

        _PHBase.__init__(self)

        self._first_solve = True

        # So we have access to real scenario and bundle probabilities
        self._uncompressed_scenario_tree = None

        # global handle to ph extension plugins
        self._ph_plugins = ExtensionPoint(IPHSolverServerExtension)
        self._modules_imported = modules_imported

    #
    # Collect full variable warmstart information off of the scenario instance
    #

    def collect_warmstart(self, scenario_name):

        if self._verbose:
            print("Received request to collect warmstart data "
                  "for scenario="+str(scenario_name))

        result = dict((symbol, vardata.value)
                      for symbol, vardata in iteritems(
                              self._instances[scenario_name].\
                              _PHInstanceSymbolMaps[Var].bySymbol))

        return result

    #
    # Overloading from _PHBase to add a few extra print statements
    #

    def activate_ph_objective_weight_terms(self):

        if self._verbose:
            print("Received request to activate PH objective weight "
                  "terms for scenario(s)="+str(list(iterkeys(self._instances))))

        if self._initialized is False:
            raise RuntimeError("PH solver server has not been initialized!")

        _PHBase.activate_ph_objective_weight_terms(self)

        if self._verbose:
            print("Activating PH objective weight terms")

    #
    # Overloading from _PHBase to add a few extra print statements
    #

    def deactivate_ph_objective_weight_terms(self):

        if self._verbose:
            print("Received request to deactivate PH objective weight terms for scenario(s)="+str(list(iterkeys(self._instances))))

        if self._initialized is False:
            raise RuntimeError("PH solver server has not been initialized!")

        _PHBase.deactivate_ph_objective_weight_terms(self)

        if self._verbose:
            print("Deactivating PH objective weight terms")

    #
    # Overloading from _PHBase to add a few extra print statements
    #

    def activate_ph_objective_proximal_terms(self):

        if self._verbose:
            print("Received request to activate PH objective proximal terms for scenario(s)="+str(list(iterkeys(self._instances))))

        if self._initialized is False:
            raise RuntimeError("PH solver server has not been initialized!")

        _PHBase.activate_ph_objective_proximal_terms(self)

        if self._verbose:
            print("Activating PH objective proximal terms")

    #
    # Overloading from _PHBase to add a few extra print statements
    #

    def deactivate_ph_objective_proximal_terms(self):

        if self._verbose:
            print("Received request to deactivate PH objective proximal terms for scenario(s)="+str(list(iterkeys(self._instances))))

        if self._initialized is False:
            raise RuntimeError("PH solver server has not been initialized!")

        _PHBase.deactivate_ph_objective_proximal_terms(self)

        if self._verbose:
            print("Deactivating PH objective proximal terms")

    def collect_scenario_tree_data(self, tree_object_names):

        data = {}
        node_data = data['nodes'] = {}
        for node_name in tree_object_names['nodes']:
            tree_node = self._scenario_tree.get_node(node_name)
            this_node_data = node_data[node_name] = {}
            this_node_data['_variable_ids'] = tree_node._variable_ids
            this_node_data['_standard_variable_ids'] = tree_node._standard_variable_ids
            this_node_data['_variable_indices'] = tree_node._variable_indices
            this_node_data['_integer'] = list(tree_node._integer)
            this_node_data['_binary'] = list(tree_node._binary)
            this_node_data['_semicontinuous'] = list(tree_node._semicontinuous)
            # master will need to reconstruct
            # _derived_variable_ids
            # _name_index_to_id

        scenario_data = data['scenarios'] = {}
        for scenario_name in tree_object_names['scenarios']:
            scenario = self._scenario_tree.get_scenario(scenario_name)
            this_scenario_data = scenario_data[scenario_name] = {}
            this_scenario_data['_objective_name'] = scenario._objective_name
            this_scenario_data['_objective_sense'] = scenario._objective_sense
            this_scenario_data['_rho'] = scenario._rho

        return data


    def initialize(self,
                   model_location,
                   data_location,
                   object_name,
                   objective_sense,
                   solver_type,
                   solver_io,
                   scenario_bundle_specification,
                   create_random_bundles,
                   scenario_tree_random_seed,
                   default_rho,
                   linearize_nonbinary_penalty_terms,
                   retain_quadratic_binary_terms,
                   breakpoint_strategy,
                   integer_tolerance,
                   output_solver_results,
                   verbose,
                   compile_scenario_instances):

        # TODO: Does this import need to be delayed because
        #       it is in a plugins subdirectory
        from pyomo.solvers.plugins.solvers.persistent_solver import \
            PersistentSolver

        if verbose:
            print("Received request to initialize PH solver server")
            print("")
            print("Model source: "+model_location)
            print("Scenario Tree source: "+str(data_location))
            print("Solver type: "+solver_type)
            print("Scenario or bundle name: "+object_name)
            if scenario_bundle_specification != None:
                print("Scenario tree bundle specification: "
                      +scenario_bundle_specification)
            if create_random_bundles != None:
                print("Create random bundles: "+str(create_random_bundles))
            if scenario_tree_random_seed != None:
                print("Scenario tree random seed: "+ str(scenario_tree_random_seed))
            print("Linearize non-binary penalty terms: "
                  + str(linearize_nonbinary_penalty_terms))

        if self._initialized:
            raise RuntimeError("PH solver servers cannot currently be "
                               "re-initialized")

        # let plugins know if they care.
        if self._verbose:
            print("Invoking pre-initialization PHSolverServer plugins")
        for plugin in self._ph_plugins:
            plugin.pre_ph_initialization(self)

        self._objective_sense = objective_sense
        self._verbose = verbose
        self._rho = default_rho
        self._linearize_nonbinary_penalty_terms = linearize_nonbinary_penalty_terms
        self._retain_quadratic_binary_terms = retain_quadratic_binary_terms
        self._breakpoint_strategy = breakpoint_strategy
        self._integer_tolerance = integer_tolerance
        self._output_solver_results = output_solver_results

        # the following is applicable to *all* solver plugins managed by this
        # server. this could be relaxed in the future, if one so desired.
        self._solver_type = solver_type
        self._solver_io = solver_io

        if self._verbose:
            print("Constructing solver type="+solver_type)

        object_solver = self._solver_map[object_name] = SolverFactory(solver_type,solver_io=self._solver_io)
        if object_solver == None:
            raise ValueError("Unknown solver type=" + solver_type + " specified")

        # we need the base model to construct
        # the scenarios that this server is responsible for.
        # TBD - revisit the various "weird" scenario tree arguments

        # GAH: At this point these should never be any form of
        #      compressed archive (unless I messed up the code) We
        #      want to avoid multiple unarchiving of the scenario and
        #      model directories. This should have been done once on
        #      the master node before this point, meaning these names
        #      should point to the unarchived directories.
        assert os.path.exists(model_location)
        assert (data_location is None) or os.path.exists(data_location)
        self._scenario_instance_factory = scenario_instance_factory = ScenarioTreeInstanceFactory(model_location,
                                                                                                  data_location)
        self._scenario_tree = scenario_instance_factory.generate_scenario_tree(
            downsample_fraction=None,
            bundles=scenario_bundle_specification,
            random_bundles=create_random_bundles,
            random_seed=scenario_tree_random_seed,
            verbose=self._verbose)

        if self._scenario_tree is None:
             raise RuntimeError("Unable to launch PH solver server - scenario tree construction failed.")

        scenarios_to_construct = []

        if self._scenario_tree.contains_bundles():

            # validate that the bundle actually exists.
            if self._scenario_tree.contains_bundle(object_name) is False:
                raise RuntimeError("Bundle="+object_name+" does not exist.")

            if self._verbose:
                print("Loading scenarios for bundle="+object_name)

            # bundling should use the local or "mini" scenario tree -
            # and then enable the flag to load all scenarios for this
            # instance.
            scenario_bundle = self._scenario_tree.get_bundle(object_name)
            scenarios_to_construct = scenario_bundle._scenario_names

        else:

            scenarios_to_construct.append(object_name)

        instance_factory = self._scenario_tree._scenario_instance_factory
        self._scenario_tree._scenario_instance_factory = None
        self._uncompressed_scenario_tree = copy.deepcopy(self._scenario_tree)
        self._scenario_tree._scenario_instance_factory = instance_factory
        # compact the scenario tree to reflect those instances for
        # which this ph solver server is responsible for constructing.
        self._scenario_tree.compress(scenarios_to_construct)

        instances = self._scenario_tree._scenario_instance_factory.\
                    construct_instances_for_scenario_tree(
                        self._scenario_tree,
                        compile_scenario_instances=compile_scenario_instances,
                        verbose=self._verbose)

        # with the scenario instances now available, have the scenario
        # tree compute the variable match indices at each node.
        self._scenario_tree.linkInInstances(instances,
                                            self._objective_sense,
                                            create_variable_ids=True)

        self._objective_sense = \
            self._scenario_tree._scenarios[0]._objective_sense

        self._setup_scenario_instances()

        # let plugins know if they care - this callback point allows
        # users to create/modify the original scenario instances
        # and/or the scenario tree prior to creating PH-related
        # parameters, variables, and the like.
        if self._verbose:
            print("Invoking post-instance-creation PHSolverServer plugins")

        for plugin in self._ph_plugins:
            plugin.post_instance_creation(self)

        # augment the instance with PH-specific parameters (weights,
        # rhos, etc).  this value and the linearization parameter as a
        # command-line argument.
        self._create_scenario_ph_parameters()

        # create symbol maps for easy storage/transmission of variable
        # values
        symbol_ctypes = (Var, Suffix)
        self._create_instance_symbol_maps(symbol_ctypes)

        # form the ph objective weight and proximal expressions Note:
        # The Expression objects for the weight and proximal terms
        # will be added to the instances objectives but will be
        # assigned values of 0.0, so that the original objective
        # function form is maintained.  The purpose is so that we can
        # use this shared Expression object in the bundle binding
        # instance objectives as well when we call
        # _form_bundle_binding_instances a few lines down (so
        # regeneration of bundle objective expressions is no longer
        # required before each iteration k solve.

        self.add_ph_objective_weight_terms()
        # Note: call this function on the base class to avoid the
        #       check for whether this phsolverserver has been
        #       initialized
        _PHBase.deactivate_ph_objective_weight_terms(self)
        self.add_ph_objective_proximal_terms()
        # Note: call this function on the base class to avoid the
        #       check for whether this phsolverserver has been
        #       initialized
        _PHBase.deactivate_ph_objective_proximal_terms(self)

        # create the bundle extensive form, if bundling.
        if self._scenario_tree.contains_bundles():
            self._form_bundle_binding_instances()

        # if we are dealing with a persisent solver plugin, go ahead
        # and compile the instance into the solver.
        # TBD - we need to pass in the symbolic solver labels to initialize()
        #       when using persistent solver plugins - because this attribute
        #       is set up-front.
        if self._verbose:
            print("Setting instance for persistent solver interface")

        # TODO: the symbolic solver label option is not propagated presently
        #       from PH to ph solver servers until the solve() method is invoked.
        #       making the default "False", as largely labels take memory and
        #       one can debug in the interim with a non-persistent solver.
        if isinstance(object_solver, PersistentSolver):
            if self._scenario_tree.contains_bundles():
                object_solver.set_instance(
                    self._bundle_binding_instance_map[object_name],
                    symbolic_solver_labels=False,
                    output_fixed_variable_bounds=self._write_fixed_variables)
            else:
                object_solver.set_instance(
                    self._scenario_tree.get_scenario(object_name).instance,
                    symbolic_solver_labels=False,
                    output_fixed_variable_bounds=self._write_fixed_variables)

        # Delay any preprocessing of the scenario instances
        # until we are inside the solve method. This gives users a
        # chance to further modify the instances (e.g., boundsetter
        # callbacks) without having to worry about setting preprocessor
        # tags because we are going to preprocess the entire instance
        # anyway.

        # let plugins know if they care.
        if self._verbose:
            print("Invoking post-initialization PHSolverServer plugins")
        for plugin in self._ph_plugins:
            plugin.post_ph_initialization(self)

        # TBD - what does initialize() even mean - we could be dealing with more than one instance.
        #       this too should probably be a map.
        # we're good to go!
        self._initialized = True

    def collect_results(self,
                        object_name,
                        results_flags):

        stages_to_load = None
        if not TransmitType.TransmitAllStages(results_flags):
            if TransmitType.TransmitNonLeafStages(results_flags):
                # exclude the leaf node
                stages_to_load = set(s.name for s in self._scenario_tree.stages[:-1])
            else:
                stages_to_load = set()

        if self._scenario_tree.contains_bundles():
            bundle = self._scenario_tree.get_bundle(object_name)
            results = {}
            for scenario_name in bundle._scenario_names:
                scenario = self._scenario_tree.get_scenario(scenario_name)
                scenario.update_solution_from_instance(stages=stages_to_load)
                results[scenario_name] = \
                    scenario.copy_solution()
        else:
            scenario = self._scenario_tree.get_scenario(object_name)
            scenario.update_solution_from_instance(stages=stages_to_load)
            results = scenario.copy_solution()

        return results

    def solve(self,
              object_name,
              tee,
              keepfiles,
              symbolic_solver_labels,
              output_fixed_variable_bounds,
              solver_options,
              solver_suffixes,
              warmstart,
              variable_transmission):
        # TODO: Does this import need to be delayed because
        #       it is in a plugins subdirectory
        from pyomo.solvers.plugins.solvers.persistent_solver import \
            PersistentSolver

        if self._initialized is False:
            raise RuntimeError("PH solver server has not been initialized!")

        # cache the type of the object, to avoid some if-thens that would
        # otherwise be necessary in the code below - largely with printing.
        if self._scenario_tree.contains_bundles():
            object_type = "bundle"
        else:
            object_type = "instance"

        if self._verbose:
            print("Received request to solve scenario %s=%s" % (object_type, object_name))

        # verify that this PH solver server actually knows about the
        # input object, and find the corresponding solver plugin.
        if self._scenario_tree.contains_bundles():
            if not self._scenario_tree.is_bundle(object_name):
                print("PH solver server requested to solve bundle=%s, "
                      " but that bundle is not recognized." % object_name)
                return None
        else:
            if not self._scenario_tree.is_scenario(object_name):
                print("PH solver server requested to solve scenario=%s, "
                      " but that scenario is not recognized." % object_name)
                return None

        if object_name not in self._solver_map:
            print("No solver plugin available for scenario%s=%s" % (object_type,object_name))
        object_solver = self._solver_map[object_name]

        self._tee = tee
        self._symbolic_solver_labels = symbolic_solver_labels
        self._write_fixed_variables = output_fixed_variable_bounds
        self._solver_options = solver_options
        self._solver_suffixes = solver_suffixes
        self._warmstart = warmstart
        self._variable_transmission = variable_transmission

        if self._first_solve:
            # let plugins know if they care.
            if self._verbose:
                print("Invoking pre-iteration-0-solve PHSolverServer plugins")
            for plugin in self._ph_plugins:
                plugin.pre_iteration_0_solve(self)
        else:
            # let plugins know if they care.
            if self._verbose:
                print("Invoking pre-iteration-k-solve PHSolverServer plugins")
            for plugin in self._ph_plugins:
                plugin.pre_iteration_k_solve(self)

        # process input solver options - they will be persistent
        # across to the next solve.  TBD: we might want to re-think a
        # reset() of the options, or something.
        for key,value in iteritems(self._solver_options):
            if self._verbose:
                print("Processing solver option="+key+", value="+str(value))
            object_solver.options[key] = value

        # with the introduction of piecewise linearization, the form
        # of the penalty-weighted objective is no longer fixed. thus,
        # when linearizing, we need to construct (or at least modify)
        # the constraints used to compute the linearized cost terms.
        if (self._linearize_nonbinary_penalty_terms > 0):
            # These functions will do nothing if ph proximal terms are
            # not present on the model
            self.form_ph_linearized_objective_constraints()
            # if linearizing, clear the values of the PHQUADPENALTY*
            # variables.  if they have values, this can intefere with
            # warm-starting due to constraint infeasibilities.
            if self._first_solve is False:
                self._reset_instance_linearization_variables()

        self._preprocess_scenario_instances()

        if not self._first_solve:

            # GAH: We may need to redefine our concept of
            #      warmstart. These values could be helpful in the
            #      nonlinear case (or could be better than 0.0, the
            #      likely default used by the solver when these
            #      initializations are not placed in the NL
            #      file. **Note: Initializations go into the NL file
            #      independent of the "warmstart" keyword

            if self._scenario_tree.contains_bundles() is True:
                # clear non-converged variables and stage cost
                # variables, to ensure feasible warm starts.
                reset_nonconverged_variables(self._scenario_tree, self._instances)
                reset_stage_cost_variables(self._scenario_tree, self._instances)
            else:
                # clear stage cost variables, to ensure feasible warm starts.
                reset_stage_cost_variables(self._scenario_tree, self._instances)

        if isinstance(object_solver, PersistentSolver):
            common_solve_kwds = {
                'load_solutions':False,
                'tee':self._tee,
                'keepfiles':keepfiles,
                'suffixes':self._solver_suffixes}
        else:
            common_solve_kwds = {
                'load_solutions':False,
                'tee':self._tee,
                'keepfiles':keepfiles,
                'symbolic_solver_labels':self._symbolic_solver_labels,
                'output_fixed_variable_bounds':self._write_fixed_variables,
                'suffixes':self._solver_suffixes}
        if object_solver.warm_start_capable():
            common_solve_kwds['warmstart'] = self._warmstart

        stages_to_load = None
        if not TransmitType.TransmitAllStages(variable_transmission):
            if TransmitType.TransmitNonLeafStages(variable_transmission):
                # exclude the leaf node
                stages_to_load = set(s.name for s in self._scenario_tree.stages[:-1])
            else:
                stages_to_load = set()

        failure = False
        results = None
        if self._scenario_tree.contains_bundles():

            bundle_ef_instance = self._bundle_binding_instance_map[object_name]

            solve_start_time = time.time()

            results = object_solver.solve(bundle_ef_instance,
                                          **common_solve_kwds)

            pyomo_solve_time = time.time() - solve_start_time

            if (len(results.solution) == 0) or \
               (results.solution(0).status == \
                SolutionStatus.infeasible) or \
               (results.solver.termination_condition == \
                TerminationCondition.infeasible):

                if self._verbose:
                    results.write()
                    print("Solve failed for bundle="
                          +object_name+"; no solutions generated")
                failure = True

            else:

                if self._verbose:
                    print("Successfully solved scenario bundle="+object_name)

                if self._output_solver_results:
                    print("Results for scenario bundle=%s:"
                          % (bundle_name))
                    print(results.write(num=1))

                # load the results into the instances on the server
                # side. this is non-trivial in terms of computation time,
                # for a number of reasons. plus, we don't want to pickle
                # and return results - rather, just variable-value maps.
                results_sm = results._smap
                bundle_ef_instance.solutions.load_from(
                    results,
                    allow_consistent_values_for_fixed_vars=\
                        self._write_fixed_variables,
                    comparison_tolerance_for_fixed_vars=\
                        self._comparison_tolerance_for_fixed_vars,
                    ignore_fixed_vars=not self._write_fixed_variables)

                if self._verbose:
                    print("Successfully loaded solution for bundle="+object_name)

                variable_values = {}
                for scenario in self._scenario_tree._scenarios:
                    scenario.update_solution_from_instance(stages=stages_to_load)
                    variable_values[scenario._name] = \
                        scenario.copy_solution()

                suffix_values = {}

                # suffixes are stored on the master block.
                bundle_ef_instance = self._bundle_binding_instance_map[object_name]

                for scenario in self._scenario_tree._scenarios:
                    # NOTE: We are only presently extracting suffix values
                    #       for constraints, as this whole interface is
                    #       experimental. And probably inefficient. But it
                    #       does work.
                    scenario_instance = self._instances[scenario._name]
                    this_scenario_suffix_values = {}
                    for suffix_name in self._solver_suffixes:
                        this_suffix_map = {}
                        suffix = getattr(bundle_ef_instance, suffix_name)
                        for constraint in scenario_instance.component_objects(Constraint, active=True):
                            this_constraint_suffix_map = {}
                            for index, constraint_data in iteritems(constraint):
                                this_constraint_suffix_map[index] = suffix.get(constraint_data)
                            this_suffix_map[constraint.name] = this_constraint_suffix_map
                        this_scenario_suffix_values[suffix_name] = this_suffix_map
                    suffix_values[scenario._name] = this_scenario_suffix_values

        else:

            scenario = self._scenario_tree._scenario_map[object_name]
            scenario_instance = self._instances[object_name]

            solve_start_time = time.time()

            results = object_solver.solve(scenario_instance,
                                          **common_solve_kwds)

            pyomo_solve_time = time.time() - solve_start_time

            if (len(results.solution) == 0) or \
               (results.solution(0).status == \
                SolutionStatus.infeasible) or \
               (results.solver.termination_condition == \
                TerminationCondition.infeasible):

                if self._verbose:
                    results.write()
                    print("Solve failed for scenario="
                          +object_name+"; no solutions generated")
                failure = True

            else:

                if self._verbose:
                    print("Successfully solved scenario instance="+object_name)

                if self._output_solver_results:
                    print("Results for scenario instance=%s:"
                          % (object_name))
                    print(results.write(num=1))

                # load the results into the instances on the server
                # side. this is non-trivial in terms of computation time,
                # for a number of reasons. plus, we don't want to pickle
                # and return results - rather, just variable-value maps.
                results_sm = results._smap
                scenario_instance.solutions.load_from(
                    results,
                    allow_consistent_values_for_fixed_vars=\
                       self._write_fixed_variables,
                    comparison_tolerance_for_fixed_vars=\
                       self._comparison_tolerance_for_fixed_vars,
                    ignore_fixed_vars=not self._write_fixed_variables)

                scenario.update_solution_from_instance(stages=stages_to_load)
                variable_values = \
                    scenario.copy_solution()

                if self._verbose:
                    print("Successfully loaded solution for scenario="+object_name)

                # extract suffixes into a dictionary, mapping suffix names
                # to dictionaries that in turn map constraint names to
                # (index, suffix-value) pairs.
                suffix_values = {}

                # NOTE: We are only presently extracting suffix values for
                #       constraints, as this whole interface is
                #       experimental. And probably inefficient. But it
                #       does work.
                for suffix_name in self._solver_suffixes:
                    this_suffix_map = {}
                    suffix = getattr(scenario_instance, suffix_name)
                    # TODO: This needs to be over all blocks
                    for constraint_name, constraint in \
                          iteritems(scenario_instance.component_map(Constraint, active=True)):
                        this_constraint_suffix_map = {}
                        for index, constraint_data in iteritems(constraint):
                            this_constraint_suffix_map[index] = \
                                suffix.get(constraint_data)
                        this_suffix_map[constraint_name] = this_constraint_suffix_map
                    suffix_values[suffix_name] = this_suffix_map

        if not failure:

            self._solver_results[object_name] = (results, results_sm)

            # auxilliary values are those associated with the solve itself.
            auxilliary_values = {}

            solution0 = results.solution(0)
            if hasattr(solution0, "gap") and \
               (not isinstance(solution0.gap, UndefinedData)) and \
               (solution0.gap is not None):
                auxilliary_values["gap"] = solution0.gap

            auxilliary_values["solve_time"], auxilliary_values["pyomo_solve_time"] = \
                extract_solve_times(results, default=None)

            auxilliary_values['solution_status'] = solution0.status.key

            solve_method_result = (variable_values, suffix_values, auxilliary_values)

        else:

            solve_method_result = ()

        if self._first_solve is True:
            # let plugins know if they care.
            if self._verbose:
                print("Invoking post-iteration-0-solve PHSolverServer plugins")
            for plugin in self._ph_plugins:
                plugin.post_iteration_0_solve(self)
        else:
            # let plugins know if they care.
            if self._verbose:
                print("Invoking post-iteration-k-solve PHSolverServer plugins")
            for plugin in self._ph_plugins:
                plugin.post_iteration_k_solve(self)

        self._first_solve = False

        return solve_method_result

    def update_xbars(self, object_name, new_xbars):

        if self._verbose:
            if self._scenario_tree.contains_bundles() is True:
                print("Received request to update xbars for bundle="+object_name)
            else:
                print("Received request to update xbars for scenario="+object_name)

        if self._initialized is False:
            raise RuntimeError("PH solver server has not been initialized!")

        for node_name, node_xbars in iteritems(new_xbars):
            tree_node = self._scenario_tree._tree_node_map[node_name]
            tree_node._xbars.update(node_xbars)

    #
    # updating weights only applies to scenarios - not bundles.
    #
    def update_weights(self, scenario_name, new_weights):

        if self._verbose:
            print("Received request to update weights for scenario="+scenario_name)

        if self._initialized is False:
            raise RuntimeError("PH solver server has not been initialized!")

        if scenario_name not in self._instances:
            print("ERROR: Received request to update weights for instance not in PH solver server instance collection!")
            return None

        scenario = self._scenario_tree._scenario_map[scenario_name]
        for tree_node_name, tree_node_weights in iteritems(new_weights):
            scenario._w[tree_node_name].update(tree_node_weights)

    #
    # updating rhos is only applicable to scenarios.
    #
    def update_rhos(self, scenario_name, new_rhos):

        if self._verbose:
            print("Received request to update rhos for scenario="+scenario_name)

        if self._initialized is False:
            raise RuntimeError("PH solver server has not been initialized!")

        if scenario_name not in self._instances:
            print("ERROR: Received request to update rhos for scenario="+scenario_name+", which is not in the PH solver server instance set="+str(self._instances.keys()))
            return None

        scenario = self._scenario_tree._scenario_map[scenario_name]
        for tree_node_name, tree_node_rhos in iteritems(new_rhos):
            scenario._rho[tree_node_name].update(tree_node_rhos)

    #
    # updating tree node statistics is bundle versus scenario agnostic.
    #

    def update_tree_node_statistics(self,
                                    scenario_name,
                                    new_node_minimums,
                                    new_node_maximums):

        if self._verbose:
            if self._scenario_tree.contains_bundles() is True:
                print("Received request to update tree node "
                      "statistics for bundle="+scenario_name)
            else:
                print("Received request to update tree node "
                      "statistics for scenario="+scenario_name)

        if self._initialized is False:
            raise RuntimeError("PH solver server has not been initialized!")

        for tree_node_name, tree_node_minimums in iteritems(new_node_minimums):
            this_tree_node_minimums = \
                self._scenario_tree._tree_node_map[tree_node_name]._minimums
            this_tree_node_minimums.update(tree_node_minimums)

        for tree_node_name, tree_node_maximums in iteritems(new_node_maximums):
            this_tree_node_maximums = \
                self._scenario_tree._tree_node_map[tree_node_name]._maximums
            this_tree_node_maximums.update(tree_node_maximums)

    #
    # define the indicated suffix on my scenario instance. not dealing
    # with bundles right now.
    #

    def define_import_suffix(self, object_name, suffix_name):

        if self._verbose:
            if self._scenario_tree.contains_bundles() is True:
                print("Received request to define import suffix="
                      +suffix_name+" for bundle="+object_name)
            else:
                print("Received request to define import suffix="
                      +suffix_name+" for scenario="+object_name)

        if self._initialized is False:
            raise RuntimeError("PH solver server has not been initialized!")

        if self._scenario_tree.contains_bundles() is True:

            bundle_ef_instance = self._bundle_binding_instance_map[object_name]

            bundle_ef_instance.add_component(suffix_name,
                                             Suffix(direction=Suffix.IMPORT))

        else:

            if object_name not in self._instances:
                print("ERROR: Received request to define import suffix="
                      +suffix_name+" for scenario="+object_name+
                      ", which is not in the collection of PH solver "
                      "server instances="+str(self._instances.keys()))
                return None
            scenario_instance = self._instances[object_name]

            scenario_instance.add_component(suffix_name,
                                            Suffix(direction=Suffix.IMPORT))

    #
    # Invoke the indicated function in the specified module.
    #

    def invoke_external_function(self,
                                 object_name,
                                 invocation_type,
                                 module_name,
                                 function_name,
                                 function_args,
                                 function_kwds):

        # pyutilib.Enum can not be serialized depending on the
        # serializer type used by Pyro, so we just send the
        # key name
        invocation_type = getattr(InvocationType,invocation_type)

        if self._verbose:
            if self._scenario_tree.contains_bundles():
                print("Received request to invoke external function"
                      "="+function_name+" in module="+module_name+" "
                      "for bundle="+object_name)
            else:
                print("Received request to invoke external function"
                      "="+function_name+" in module="+module_name+" "
                      "for scenario="+object_name)

        if self._initialized is False:
            raise RuntimeError("PH solver server has not been initialized!")

        scenario_tree_object = None
        if self._scenario_tree.contains_bundles():
            scenario_tree_object = self._scenario_tree._scenario_bundle_map[object_name]
        else:
            scenario_tree_object = self._scenario_tree._scenario_map[object_name]

        if module_name in self._modules_imported:
            this_module = self._modules_imported[module_name]
        elif module_name in sys.modules:
            this_module = sys.modules[module_name]
        else:
            this_module = pyutilib.misc.import_file(module_name,
                                                    clear_cache=True)
            self._modules_imported[module_name] = this_module

        module_attrname = function_name
        subname = None
        if not hasattr(this_module, module_attrname):
            if "." in module_attrname:
                module_attrname, subname = function_name.split(".",1)
            if not hasattr(this_module, module_attrname):
                raise RuntimeError("Function="+function_name+" is not present "
                                   "in module="+module_name)

        if function_args is None:
            function_args = ()
        if function_kwds is None:
            function_kwds = {}

        call_objects = None
        if invocation_type == InvocationType.SingleInvocation:
            if self._scenario_tree.contains_bundles():
                call_objects = (object_name,self._scenario_tree._scenario_bundle_map[object_name])
            else:
                call_objects = (object_name,self._scenario_tree._scenario_map[object_name])
        elif (invocation_type == InvocationType.PerBundleInvocation) or \
             (invocation_type == InvocationType.PerBundleChainedInvocation):
            if not self._scenario_tree.contains_bundles():
                raise ValueError("Received request for bundle invocation type "
                                 "but the scenario tree does not contain bundles.")
            call_objects = iteritems(self._scenario_tree._scenario_bundle_map)
        elif (invocation_type == InvocationType.PerScenarioInvocation) or \
             (invocation_type == InvocationType.PerScenarioChainedInvocation):
            call_objects = iteritems(self._scenario_tree._scenario_map)
        elif (invocation_type == InvocationType.PerNodeInvocation) or \
             (invocation_type == InvocationType.PerNodeChainedInvocation):
            call_objects = iteritems(self._scenario_tree._tree_node_map)
        else:
            raise ValueError("Unexpected function invocation type '%s'. "
                             "Expected one of %s"
                             % (invocation_type,
                                [str(v) for v in InvocationType._values]))

        function = getattr(this_module, module_attrname)
        if subname is not None:
            function = getattr(function, subname)

        if invocation_type == InvocationType.SingleInvocation:
            call_name, call_object = call_objects
            return function(self,
                            self._scenario_tree,
                            call_object,
                            *function_args,
                            **function_kwds)
        elif (invocation_type == InvocationType.PerBundleChainedInvocation) or \
             (invocation_type == InvocationType.PerScenarioChainedInvocation) or \
             (invocation_type == InvocationType.PerNodeChainedInvocation):
            result = function_args
            for call_name, call_object in call_objects:
                result = function(self,
                                  self._scenario_tree,
                                  call_object,
                                  *result,
                                  **function_kwds)
            return result
        else:
            return dict((call_name,function(self,
                                            self._scenario_tree,
                                            call_object,
                                            *function_args,
                                            **function_kwds))
                        for call_name, call_object in call_objects)

    #
    # restore solutions for all of scenario instances.
    #

    def restoreCachedSolutions(self, object_name, cache_id, release_cache):

        if self._verbose:
            if self._scenario_tree.contains_bundles() is True:
                print("Received request to restore cached solution for bundle="+object_name)
            else:
                print("Received request to restore cached solution for scenario="+object_name)
            print("Restoring from cache id: "+str(cache_id))

        if self._initialized is False:
            raise RuntimeError("PH solver server has not been initialized!")

        if self._scenario_tree.contains_bundles() is True:
            # validate that the bundle actually exists.
            if self._scenario_tree.contains_bundle(object_name) is False:
                raise RuntimeError("Bundle="+object_name+" does not exist.")
        else:
            if object_name not in self._instances:
                raise RuntimeError(object_name+" is not in the PH solver server instance collection")

        _PHBase.restoreCachedSolutions(self, cache_id, release_cache)

    #
    # restore solutions for all of scenario instances.
    #

    def cacheSolutions(self, object_name, cache_id):

        if self._verbose:
            if self._scenario_tree.contains_bundles() is True:
                print("Received request to cache solution for bundle="+object_name)
            else:
                print("Received request to cache solution for scenario="+object_name)
            print("Caching with id: "+str(cache_id))

        if self._initialized is False:
            raise RuntimeError("PH solver server has not been initialized!")

        if self._scenario_tree.contains_bundles() is True:
            # validate that the bundle actually exists.
            if self._scenario_tree.contains_bundle(object_name) is False:
                raise RuntimeError("Bundle="+object_name+" does not exist.")
        else:
            if object_name not in self._instances:
                raise RuntimeError(object_name+" is not in the PH solver server instance collection")

        _PHBase.cacheSolutions(self, cache_id)


    #
    # fix variables as instructed by the PH client.
    #
    def update_fixed_variables(self, object_name, fixed_variables):

        if self._verbose:
            if self._scenario_tree.contains_bundles() is True:
                print("Received request to update fixed variables for "
                      "bundle="+object_name)
            else:
                print("Received request to update fixed variables for "
                      "scenario="+object_name)

        if self._initialized is False:
            raise RuntimeError("PH solver server has not been initialized!")

        for node_name, node_fixed_vars in iteritems(fixed_variables):
            tree_node = self._scenario_tree.get_node(node_name)
            tree_node._fix_queue.update(node_fixed_vars)

        self._push_fix_queue_to_instances()

    def process(self, data):

        result = None
        if data.action == "initialize":
            result = self.initialize(data.model_location,
                                     data.data_location,
                                     data.object_name,
                                     data.objective_sense,
                                     data.solver_type,
                                     data.solver_io,
                                     data.scenario_bundle_specification,
                                     data.create_random_bundles,
                                     data.scenario_tree_random_seed,
                                     data.default_rho,
                                     data.linearize_nonbinary_penalty_terms,
                                     data.retain_quadratic_binary_terms,
                                     data.breakpoint_strategy,
                                     data.integer_tolerance,
                                     data.output_solver_results,
                                     data.verbose,
                                     data.compile_scenario_instances)

        elif data.action == "collect_results":
            result = self.collect_results(data.name,
                                          data.var_config)
        elif data.action == "solve":
            # we are adding the following code because some solvers, including
            # CPLEX, are not all that robust - in that they can spontaneously
            # and sporadically fail. ultimately, this should be command-line
            # option driver.
            max_num_attempts = 2
            attempts_so_far = 0
            successful_solve = False
            while (not successful_solve):
                try:
                    attempts_so_far += 1
                    result = self.solve(data.name,
                                        data.tee,
                                        data.keepfiles,
                                        data.symbolic_solver_labels,
                                        data.output_fixed_variable_bounds,
                                        data.solver_options,
                                        data.solver_suffixes,
                                        data.warmstart,
                                        data.variable_transmission)
                    successful_solve = True
                except pyutilib.common.ApplicationError as exc:
                    print("Solve failed for object=%s - this was attempt=%d"
                          % (data.name, attempts_so_far))
                    if (attempts_so_far == max_num_attempts):
                        print("Aborting PH solver server - the maximum number "
                              "of solve attempts=%d have been executed"
                              % (max_num_attempts))
                        raise exc

            if attempts_so_far > 1:
                print("Successfully recovered from failed solve for "
                      "object=%s" % (data.name))

        elif data.action == "activate_ph_objective_proximal_terms":
            self.activate_ph_objective_proximal_terms()
            result = True

        elif data.action == "deactivate_ph_objective_proximal_terms":
            self.deactivate_ph_objective_proximal_terms()
            result = True

        elif data.action == "activate_ph_objective_weight_terms":
            self.activate_ph_objective_weight_terms()
            result = True

        elif data.action == "deactivate_ph_objective_weight_terms":
            self.deactivate_ph_objective_weight_terms()
            result = True

        elif data.action == "load_rhos":
            if self._scenario_tree.contains_bundles() is True:
                for scenario_name, scenario_instance in iteritems(self._instances):
                    self.update_rhos(scenario_name,
                                     data.new_rhos[scenario_name])
            else:
                self.update_rhos(data.name,
                                 data.new_rhos)
            result = True
            self._push_rho_to_instances()

        elif data.action == "update_fixed_variables":
            self.update_fixed_variables(data.name,
                                        data.fixed_variables)
            result = True
            self._push_fix_queue_to_instances()

        elif data.action == "load_weights":
            if self._scenario_tree.contains_bundles() is True:
                for scenario_name, scenario_instance in iteritems(self._instances):
                    self.update_weights(scenario_name,
                                        data.new_weights[scenario_name])
            else:
                self.update_weights(data.name,
                                    data.new_weights)
            result = True
            self._push_w_to_instances()

        elif data.action == "load_xbars":
            self.update_xbars(data.name,
                              data.new_xbars)
            result = True
            self._push_xbar_to_instances()

        elif data.action == "load_tree_node_stats":
            self.update_tree_node_statistics(data.name,
                                             data.new_mins,
                                             data.new_maxs)
            result = True

        elif data.action == "define_import_suffix":
            self.define_import_suffix(data.name,
                                      data.suffix_name)
            result = True

        elif data.action == "invoke_external_function":
           result = self.invoke_external_function(data.name,
                                                  data.invocation_type,
                                                  data.module_name,
                                                  data.function_name,
                                                  data.function_args,
                                                  data.function_kwds)

        elif data.action == "restore_cached_scenario_solutions":
            # don't pass the scenario argument - by default,
            # we restore solutions for all of our instances.
            self.restoreCachedSolutions(data.name,
                                        data.cache_id,
                                        data.release_cache)
            result = True
        elif data.action == "cache_scenario_solutions":
            # don't pass the scenario argument - by default,
            # we restore solutions for all of our instances.
            self.cacheSolutions(data.name,
                                data.cache_id)
            result = True
        elif data.action == "collect_scenario_tree_data":
            result = self.collect_scenario_tree_data(data.tree_object_names)
        elif data.action == "collect_warmstart":
            result = self.collect_warmstart(data.scenario_name)
        else:
            raise RuntimeError("ERROR: Unknown action="+str(data.action)+" received by PH solver server")

        return result
#
# utility method to construct an option parser for ph arguments, to be
# supplied as an argument to the runph method.
#

def construct_options_parser(usage_string):

    parser = OptionParser()
    parser.add_option("--verbose",
                      help="Generate verbose output for both initialization and execution. Default is False.",
                      action="store_true",
                      dest="verbose",
                      default=False)
    parser.add_option("--profile",
                      help="Enable profiling of Python code.  The value of this option is the number of functions that are summarized.",
                      action="store",
                      dest="profile",
                      type="int",
                      default=0)
    parser.add_option("--disable-gc",
                      help="Disable the python garbage collecter. Default is False.",
                      action="store_true",
                      dest="disable_gc",
                      default=False)
    parser.add_option('--traceback',
                      help="When an exception is thrown, show the entire call stack. Ignored if profiling is enabled. Default is False.",
                      action="store_true",
                      dest="traceback",
                      default=False)
    parser.add_option('--user-defined-extension',
                      help="The name of a python module specifying a user-defined PHSolverServer extension plugin.",
                      action="append",
                      dest="user_defined_extensions",
                      type="string",
                      default=[])
    parser.add_option('--pyro-host',
      help="The hostname to bind on when searching for a Pyro nameserver.",
      action="store",
      dest="pyro_host",
      default=None)
    parser.add_option('--pyro-port',
      help="The port to bind on when searching for a Pyro nameserver.",
      action="store",
      dest="pyro_port",
      type="int",
      default=None)

    #parser.add_option("--shutdown-on-error",
    #                  help="On error, shut down all Pyro-related components connected to the current nameserver. In most cases, it is only necessary to supply this option to the runph command.",
    #                  action="store_true",
    #                  dest="shutdown_on_error",
    #                  default=False)

    parser.usage=usage_string

    return parser

#
# Execute the PH solver server daemon.
#
def exec_phsolverserver(options):

    # disable all plugins up-front. then, enable them on an as-needed
    # basis later in this function.
    ph_extension_point = ExtensionPoint(IPHSolverServerExtension)

    for plugin in ph_extension_point:
        plugin.disable()

    if len(options.user_defined_extensions) > 0:
        for this_extension in options.user_defined_extensions:
            if this_extension in sys.modules:
                print("User-defined PHSolverServer extension module="
                      +this_extension+" already imported - skipping")
            else:
                print("Trying to import user-defined PHSolverServer "
                      "extension module="+this_extension)
                # make sure "." is in the PATH.
                original_path = list(sys.path)
                sys.path.insert(0,'.')
                pyutilib.misc.import_file(this_extension)
                print("Module successfully loaded")
                sys.path[:] = original_path # restore to what it was

            # now that we're sure the module is loaded, re-enable this
            # specific plugin.  recall that all plugins are disabled
            # by default in phinit.py, for various reasons. if we want
            # them to be picked up, we need to enable them explicitly.
            import inspect
            module_to_find = this_extension
            if module_to_find.rfind(".py"):
                module_to_find = module_to_find.rstrip(".py")
            if module_to_find.find("/") != -1:
                module_to_find = string.split(module_to_find,"/")[-1]

            for name, obj in inspect.getmembers(sys.modules[module_to_find], inspect.isclass):
                # the second condition gets around goofyness related to issubclass returning
                # True when the obj is the same as the test class.
                if issubclass(obj, SingletonPlugin) and name != "SingletonPlugin":
                    ph_extension_point = ExtensionPoint(IPHSolverServerExtension)
                    for plugin in ph_extension_point(all=True):
                        if isinstance(plugin, obj):
                            plugin.enable()

    try:
        # spawn the daemon
        TaskWorkerServer(PHPyroWorker,
                         host=options.pyro_host,
                         port=options.pyro_port)
    except:
        # if an exception occurred, then we probably want to shut down
        # all Pyro components.  otherwise, the PH client may have
        # forever while waiting for results that will never
        # arrive. there are better ways to handle this at the PH
        # client level, but until those are implemented, this will
        # suffice for cleanup.
        #NOTE: this should perhaps be command-line driven, so it can
        #      be disabled if desired.
        print("PH solver server aborted. Sending shutdown request.")
        shutdown_pyro_components(host=options.pyro_host,
                                 port=options.pyro_port,
                                 num_retries=0)
        raise

@pyomo_command('phsolverserver', "Pyro-based server for PH solvers")
def main(args=None):
    #
    # Top-level command that executes the ph solver server daemon.
    #

    #
    # Import plugins
    #
    import pyomo.environ

    #
    # Parse command-line options.
    #
    try:
        options_parser = \
            construct_options_parser("phsolverserver [options]")
        (options, args) = options_parser.parse_args(args=args)
    except SystemExit as _exc:
        # the parser throws a system exit if "-h" is specified
        # - catch it to exit gracefully.
        return _exc.code

    return launch_command(exec_phsolverserver,
                          options,
                          error_label="phsolverserver: ",
                          disable_gc=options.disable_gc,
                          profile_count=options.profile,
                          traceback=options.traceback)



#
# a model that is missing a declaration of
# a stochastic objective constant
#
from pyomo.pysp.tests.convert.utils import *

pysp_scenario_tree_model_callback = \
    simple_twostage_scenario_tree

def pysp_instance_creation_callback(scenario_name, node_names):
    model = simple_twostage_model()
    model.sobj = StochasticObjectiveAnnotation()
    model.sobj.declare(model.o, include_constant=False)
    if scenario_name == "Scenario1":
        model.o.expr = model.o.expr + 0
    elif scenario_name == "Scenario2":
        model.o.expr = model.o.expr + 1
    else:
        assert False
    return model
